import os
import json
import copy
import re
import logging
import traceback
import importlib
import subprocess
import base64
import inspect
from datetime import datetime
from dateutil import parser as dateparser
from dateutil.tz import tzlocal
from colored import fg, bg, attr
from cryptography.fernet import Fernet

from .exceptions import *
from .context import *
from .resources import *

logger = logging.getLogger("app")

def greenbg(txt):
    return "{} \u2713 {} {}".format(bg("light_green"), txt, attr(0))


def redbg(txt):
    return "{} \u274C {} {}".format(bg("indian_red_1a"), txt, attr(0))


def orangebg(txt):
    return "{} \u0023 {} {}".format(bg("dark_orange_3a"), txt, attr(0))


cached_siteconf = {}
cached_versionmap = {}

# Have we already warned the user about over-rides
warned_overrides = False

def write_encypted_creds(key, creds, encfile=None):

    key = base64.urlsafe_b64encode(key[:32].encode("utf-8"))
    f = Fernet(key)

    # Turn into text
    content = json.dumps(creds)

    # Encrypt
    contentb= f.encrypt(content.encode("utf-8"))

    # Where should I put this?
    if encfile is None:
        etcdir = os.environ.get("ENRICH_ETC", os.path.join(os.environ.get("ENRICH_ROOT"), "etc"))
        filename = os.path.join(etcdir, "sitecred.enc")
    else:
        filename = encfile

    with open(filename, "wb") as fd:
        fd.write(contentb)

def read_encrypted_creds(key, encfile=None):

    key = base64.urlsafe_b64encode(key[:32].encode("utf-8"))
    f = Fernet(key)

    # Where should I put this?
    if encfile is None:
        etcdir = os.environ.get("ENRICH_ETC", os.path.join(os.environ.get("ENRICH_ROOT"), "etc"))
        filename = os.path.join(etcdir, "sitecred.enc")
    else:
        filename = encfile

    if not os.path.exists(filename):
        return {}

    try:
        with open(filename, "rb") as fd:
            contentb = fd.read()
            content = f.decrypt(contentb)
            params = json.loads(content)
    except:
        logger.error("Failed to read sitecred.enc")
        params = {}

    return params

def read_siteconf_json(filename, app={}, context={}):

    global cached_siteconf

    try:
        # Try the sitelib first.
        import sitelib

        # Return cached file
        if filename in cached_siteconf:
            return cached_siteconf[filename]

        nature, siteconf = sitelib.read_siteconf(filename, app)
        if siteconf is None:
            raise Exception("Failed to read siteconf")

        # Cache it
        cached_siteconf[filename] = siteconf
        if nature == "open":
            logger.warning("Successfully read siteconf but it is insecure")
        else:
            logger.warning("Successfully read secure siteconf")
    except:
        # sitelib may not be installed
        # logger.debug("Could not read secure siteconf. Falling back")
        try:
            # If this is a jsonx file, then try the json version of
            # the file.
            if filename.endswith("jsonx"):
                filename = filename.replace("jsonx", "json")
            siteconf = json.load(open(filename))
            cached_siteconf[filename] = siteconf
        except:
            traceback.print_exc()
            siteconf = None

    if siteconf is None:
        raise Exception("Could not read siteconf")

    if not isinstance(siteconf, dict):
        raise Exception(
            "Invalid siteconf found. Expecting 'dict' found '{}'".format(
                str(type(siteconf))
            )
        )

    return siteconf


def read_siteconf_obs(filename=None, app={}, context={}):

    # If context not available, then generate it.
    if len(context) == 0:
        context = Context().asdict()

    # This overrides everything else...
    if "ENRICH_SITECONF" in os.environ:
        try:
            import sitelib

            encoded = os.environ["ENRICH_SITECONF"]
            confbytes = base64.b64decode(encoded)
            siteconf = sitelib.decode_bytes(confbytes)
            siteconf = json.loads(siteconf)
            return siteconf
        except:
            pass

    if filename is not None:
        return read_siteconf_json(filename, app)

    tries = [
        "siteconf.jsonx",
        "siteconf.json",
        "etc/siteconf.jsonx",
        "etc/siteconf.json",
        "enrich/etc/siteconf.jsonx",
        "enrich/etc/siteconf.json"
    ]
    try:
        tries.extend(
            [
                "%(ENRICH_ETC)s/siteconf.jsonx" % context,
                "%(ENRICH_ETC)s/siteconf.json" % context,
            ]
        )
    except:
        pass

    tries.extend(
        [
            os.path.expandvars("${ENRICH_ETC}/siteconf.jsonx"),
            os.path.expandvars("${ENRICH_ETC}/siteconf.json"),
        ]
    )

    siteconf = None
    for filename in tries:
        if os.path.exists(filename):
            siteconf = read_siteconf_json(filename, app, context)
            if siteconf is None:
                continue
            else:
                break

    if siteconf is None:
        logger.warning(
            "Could not find siteconf. Using empty dict"
        )

        siteconf = {
            'credentials': {}
        }

    # Now load the encrypted

    return siteconf

def read_siteconf(filename=None, app={}, context={}):

    # Read the default
    siteconf = read_siteconf_obs(filename, app, context)

    if (("DJANGO_SECRET_KEY" not in os.environ) and
        ("SECRET_KEY" not in os.environ)):
        return siteconf

    try:

        if "SECRET_KEY" in os.environ:
            key = os.environ['SECRET_KEY']
        else:
            key = os.environ['DJANGO_SECRET_KEY']

        encsiteconf = read_encrypted_creds(key)

        # => First incorporate the credentials
        if 'credentials' in encsiteconf:
            creds = encsiteconf['credentials']
        else:
            creds = encsiteconf
        default = siteconf.pop('credentials', {})
        overrides = []
        for k, v in creds.items():
            if k in default:
                overrides.append(k)
            default[k] = v
        siteconf['credentials'] = default

        # Next include the timezone and such
        if 'timezone' in encsiteconf:
            siteconf['timezone'] = encsiteconf['timezone']
            overrides.append('timezone')

        if 'dashboard' in encsiteconf:
            for k, v in encsiteconf['dashboard'].items():
                siteconf['dashboard'][k] = v
                overrides.append(k)

        if 'envvars' in encsiteconf:
            encvariables = encsiteconf['envvars']
            if not isinstance(encvariables, dict):
                encvariables = {}
            variables = siteconf.get('variables', {})
            for k, v in encvariables.items():
                variables[k] = v
                overrides.append(k)
            siteconf['variables'] = variables

        # Note any overrides
        if len(overrides) > 0:
            global warned_overrides
            if not warned_overrides:
                logger.warning("Overriding credentials from sitecred",
                               extra={
                                   'data': ", ".join(overrides)
                               })
                warned_overrides = True

    except:
        traceback.print_exc()
        logger.error("Error while reading sitecred")

    return siteconf


def get_credentials(siteconf, name):
    """
    Look up the credentials file
    """

    if (siteconf is None) or (not isinstance(siteconf, dict)):
        raise Exception("Invalid siteconf - None or has invalid datatype")

    if ("credentials" not in siteconf) or (name not in siteconf["credentials"]):
        raise Exception("missing credentials")

    return siteconf["credentials"][name]


def get_credentials_by_name(name, app={}, context={}):
    """
    Lookup credentials by name
    """

    siteconf = read_siteconf(app=app, context={})
    return get_credentials(siteconf, name)


def get_credentials_by_type(conditions, app={}, context={}):
    """
    Lookup credentials by type
    """

    siteconf = read_siteconf(app=app, context={})

    if (siteconf is None) or (not isinstance(siteconf, dict)):
        raise Exception("Invalid siteconf - None or has invalid datatype")

    if "credentials" not in siteconf:
        raise Exception("missing credentials")

    if not isinstance(conditions, dict):
        raise Exception("Invalid condition specification")

    final = {}
    for name, detail in siteconf["credentials"].items():

        if not isinstance(detail, dict):
            continue

        if not detail.get("enable", True):
            continue

        match = True
        for var, choices in conditions.items():
            if var not in detail:
                match = False
                break
            if detail[var] not in choices:
                match = False
                break

        if not match:
            continue

        final[name] = copy.deepcopy(detail)

    return final

############################################
# Access emails
############################################
def get_siteconf_var(var, section=None, default=None):

    siteconf = read_siteconf()

    if ((section in siteconf) and
        (isinstance(siteconf[section], dict))):
        return siteconf[section].get(var, default)

    return siteconf.get(var, default)

def get_env_var(var, default=None, debug=False):
    """
    Get variable from environment. Account for
    """
    varlist = []
    if isinstance(var, list):
        varlist = var
    else:
        varlist = [str(var)]

    if debug:
        print("varlist", varlist)

    siteconf = read_siteconf()

    if debug:
        print(siteconf.get('variables', {}))
    variables = siteconf.get('variables', {})
    for var in varlist:
        if var in variables:
            if debug:
                print("Found in siteconf", var, "with value", variables[var])
            return variables[var]

    for var in varlist:
        if var in os.environ:
            if debug:
                print("Found in env", var, "with value", os.environ[var])
            return os.environ[var]

    return default

def get_env_var_list(var, separator=",", default=[]):

    value = get_env_var(var)
    if not isinstance(value, str):
        return default

    values = value.split(separator)
    values = [v for v in values if len(v) > 0]
    return values

def get_env_var_bool(var, default=False):

    value = get_env_var(var)
    if not isinstance(value, str):
        return default

    return value.lower().strip() == 'true'

def get_default_from_email(name='from'):
    """
    Get default emails
    """

    siteconf = read_siteconf()
    notification = siteconf.get('notification', {})
    if name in notification:
        return notification[name]

    email = get_env_var(['DJANGO_DEFAULT_FROM_EMAIL',
                        'DEFAULT_FROM_EMAIL'],
                        default='Scribble Support<support@scribbledata.io>')

    return email

def update_from_git(detail, context, debug=False):

    # Make a local version
    context = copy.copy(context)

    label = detail.get("label", detail.get("repo", "unknown"))

    # Try the detail run root
    root = detail["git"] % context
    if not os.path.exists(root):

        # try one alternative path if possible
        if "ENRICH_RUN_ROOT" not in detail["git"]:
            raise Exception("Expected ENRICH_RUN_ROOT: {}".format(detail["git"]))

        # Fallback to the default run_root
        context["ENRICH_RUN_ROOT"] = context["ENRICH_ROOT"]
        altroot = detail["git"] % context
        if not os.path.exists(altroot):
            raise Exception("Missing repo roots: {}, {}".format(root, altroot))
        root = altroot

    # Update in place
    detail["git"] = root

    # By default use master branch only..
    branch = detail.get("branch", "master")

    # Collect all the notes
    cmd = "git -C {} log {} --pretty='%h--%ci--%N'".format(root, branch)
    notes = subprocess.check_output(cmd, shell=True)
    notes = notes.decode("utf-8")
    notes = notes.split("\n")
    notes = [
        n.split("--") for n in notes if (n.count("--") == 2 and not n.endswith("--"))
    ]  # remove empty

    detail["notes"] = [
        {"commit": n[0], "commit_date": n[1], "message": n[2]} for n in notes
    ]

    # => Collect branches..
    cmd = """git -C {} branch -l""".format(root)
    lines = subprocess.check_output(cmd, shell=True)
    lines = lines.decode("utf-8")
    lines = lines.split("\n")
    lines = [l[2:] for l in lines]
    detail["branches"] = lines

    # Collect the git commits
    cmd = """git -C {} log {} --no-walk --tags --all --pretty="%h\t%D\t%ce\t%ci\t%s" --decorate=full --date=short""".format(
        root, branch
    )
    lines = subprocess.check_output(cmd, shell=True)
    lines = lines.decode("utf-8")
    lines = lines.split("\n")
    lines = [l.strip() for l in lines]
    lines = [l.split("\t") for l in lines]
    lines = [l for l in lines if len(l) == 5]
    alltags = []
    for l in lines:

        if debug:
            print(l)

        # Skip all irrelevant remotes except upstream for the given branch
        refs = [x.strip() for x in l[1].split(",")]

        # Use relevant records. There is stuff from other branches,
        valid = False
        for r in refs:
            if (
                ("refs/heads/{}".format(branch) in r)
                or ("refs/remotes/origin/{}".format(branch) in r)
                or ("refs/tags" in r)
                or ("HEAD" in r)
            ):
                if debug:
                    print("Match", r)
                valid = True
                break

        if not valid:
            if debug:
                print("Skipping")
            continue

        # Is this head?
        head = any([ref.startswith("HEAD") for ref in refs])

        # Align the local date
        try:
            dt = dateparser.parse(l[3]).astimezone(tzlocal()).isoformat()
        except:
            dt = l[3]

        entry = {
            # ['416366c', 'tag: refs/tags/v1.4.0, refs/heads/master', '2020-03-03', 'pingali@gmail.com', 'Bug']
            "commit": l[0],
            "refs": refs,
            "author": l[2],
            "date": dt,
            "log": l[4],
        }

        # Extract the tag for this commit
        for ref in refs:
            match = re.search(r"tag: refs/tags/(\S+)", ref)
            if match is not None:
                tag = match.group(1)
                entry["tag"] = tag
                break

        if "tag" not in entry:
            entry["tag"] = l[0]

        if debug:
            print(entry)

        alltags.append(entry)

    # Now sort by reverse commit date
    alltags = sorted(alltags, key=lambda entry: entry["date"], reverse=True)
    detail["alltags"] = alltags

    position = "post"
    for entry in alltags:

        head = any([ref.startswith("HEAD") for ref in entry["refs"]])

        # Where is this position...
        if head:
            entry["position"] = "head"
            position = "pre"
        else:
            entry["position"] = position

        if head:
            detail["release"] = entry["tag"]
            detail["commit"] = entry["commit"]
            detail["date"] = entry["date"]

            # Insert the URL as well
            if (("url" in detail) and ("tree" in detail['url'])):
                urlparts = detail["url"].split("/")
                urlparts[-1] = detail["commit"]
                url = "/".join(urlparts)
                detail["url"] = url


def update_from_package(detail, context):

    pkgname = detail["package"]
    mod = importlib.import_module(pkgname)

    # v0.4.3+0.g659ce50.dirty
    if not hasattr(mod, "__version_detailed__"):
        return

    pkgdetails = mod.__version_detailed__
    release = pkgdetails["version"]
    commit = pkgdetails["full-revisionid"]

    # Now overwrite the release information
    detail["release"] = release
    detail["commit"] = commit[:8]
    if "url" in detail:
        urlparts = detail["url"].split("/")
        urlparts[-1] = detail["commit"]
        url = "/".join(urlparts)
        detail["url"] = url
    else:
        detail["release"] += " (out of date; version missing)"
        logger.error(
            "VersionMap: Cannot collect version information for {}".format(pkgname),
            extra={"data": json.dumps(detail, indent=4)},
        )

def get_versionmap_from_subdir():

    versionmap = []

    # Check if there are repos that need to be included
    etcdir = os.environ.get("ENRICH_ETC",
                            os.path.join(os.environ.get('ENRICH_ROOT'),
                                         'etc'))
    versiondir = os.path.join(etcdir, "versions")
    if not os.path.exists(versiondir):
        return versionmap

    files = os.listdir(versiondir)
    for f in files:
        fullpath = os.path.join(versiondir, f)
        try:
            versionmap.append(json.load(open(fullpath)))
        except:
            continue

    return versionmap

def get_versionmap_from_code():

    versionmap = []

    try:
        from git import Repo

        # Check if there are repos that need to be included
        customerdir = os.environ.get("ENRICH_CUSTOMERS",
                                     os.path.join(os.environ.get('ENRICH_ROOT'),
                                                  'customers'))
        for c in os.listdir(customerdir):
            repodir = os.path.join(customerdir, c)
            try:
                if not os.path.exists(os.path.join(repodir, ".git")):
                    continue

                repo = Repo(repodir)
                reader = repo.config_reader()
                if not reader.has_section('remote "origin"'):
                    continue
                repodata = dict(reader.items('remote "origin"'))
                {
                    "date": "2019-12-11 15:33:47 +0530",
                    "label": "scribble",
                    "description": "Core scribble applications",
                    "repo": "enrich-scribble",
                    "dynamic": True,
                    "git": "%(ENRICH_RUN_ROOT)s/customers/scribble",
                    "commit": "f58a048d380c911fd2d7bd8fa5909499f8d36c25",
                    "notes": [],
                    "url": "https://github.com/pingali/enrich-scribble/tree/f58a048d380c911fd2d7bd8fa5909499f8d36c25",
                    "release": "v1.0.9"
                },
                branch = repo.active_branch.name
                url = repodata['url']
                if (("https://" in url) and ("@" in url)):
                    url = "https://xxx" + url[url.index("@"):]
                entry = {
                    "label": c,
                    "repo": c,
                    "description": "Unspecified",
                    "branch": branch,
                    "url": url,
                    "git": "%(ENRICH_RUN_ROOT)s/customers/" + c,
                    "dynamic": True,
                }
                versionmap.append(entry)

            except:
                #traceback.print_exc()
                pass
    except:
        pass

    return versionmap

def read_versionmap(filename=None, include_notes=False,
                    include_tags=False, context={},
                    ignore_cache=False):

    global cached_versionmap

    if isinstance(context, dict):
        if len(context) == 0:
            context = Context().asdict()
    elif isinstance(context, Context):
        context = context.asdict()

    default = [
        {
            "repo": "error",
            "release": "v1xxx",
            "date": "2018-10-09 12:06:40 +0530",
            "commit": "5079404f12df939c138215ed1a02e4d0d8814e00",
            "description": "Error while reading the versionmap",
            "url": "https://github.com/pingali/error/commit/5079404f12df939c138215ed1a02e4d0d8814e00",
            "label": "error",
            "notes": [
                {
                    "commit": "6271dcae",
                    "commit_date": "2018-10-09 12:06:40 +0530",
                    "notes": "v0 of the release",
                }
            ],
        }
    ]

    tries = [
        "versionmap.json",
        "etc/versionmap.json",
        "enrich/etc/versionmap.json",
        context.get('versionmap', None),
    ]
    if filename is not None:
        tries.append(filename)
    if "ENRICH_ETC" in context:
        altfilename = "%(ENRICH_ETC)s/versionmap.json" % context
        tries.append(altfilename)
    elif "ENRICH_ETC" in os.environ:
        altfilename = os.path.expandvars("$ENRICH_ETC/versionmap.json")
        tries.append(altfilename)

    selected_filename = None
    for f in tries:
        if ((f is not None) and (os.path.exists(f))):
            selected_filename = f

    now = datetime.now().replace(microsecond=0).isoformat()
    if selected_filename is None:
        error = "Missing versionmap: {}".format(filename)
        logger.error(error)
        default[0]["description"] = error
        return None, default

    # Make the selected filename as default
    filename = selected_filename

    # => use the filename's timestamp as the default timestamp
    timestamp = datetime.fromtimestamp(os.path.getmtime(filename))
    timestamp = timestamp.isoformat()

    if ((filename in cached_versionmap) and (not ignore_cache)):
        return timestamp, cached_versionmap[filename]

    try:
        versionmap = json.load(open(filename))
    except:
        error = "Invalid version file. Not a valid json file: {}".format(filename)
        logger.error(error)
        default[0]["description"] = error
        return default

    # => Newer dict format of the versionmap
    if isinstance(versionmap, dict):
        timestamp = versionmap.get("timestamp", timestamp)
        versionmap = versionmap.get("versionmap", None)

    if not isinstance(versionmap, list):
        error = "Invalid version map. Not a list: {}".format(filename)
        logger.error(error)
        default[0]["description"] = error
        return timestamp, default

    # Insert information from the source repos
    versionmapextra = get_versionmap_from_code()
    for extrav in versionmapextra:
        label = extrav['label']
        found = False
        for v in versionmap:
            if v.get('label', 'unknown') == label:
                found = True
                break
        if found:
            continue
        versionmap.append(extrav)

    #
    versionmapextra = get_versionmap_from_subdir()
    for extrav in versionmapextra:
        label = extrav['label']
        # Replace existing entry with the modified entry
        replaced = False
        cleaned_versionmap = []
        for v in versionmap:
            if v.get('label', 'unknown') == label:
                replaced = True
                cleaned_versionmap.append(extrav)
            else:
                cleaned_versionmap.append(v)

        versionmap = cleaned_versionmap


    # Update the release and commit information if dynamic modules are
    # specified.
    for i in range(len(versionmap)):

        label = versionmap[i].get("label", versionmap[i].get("repo", "unknown"))

        if "branch" not in versionmap[i]:
            versionmap[i]["branch"] = "master"

        if versionmap[i].get("dynamic", False):
            try:
                if "package" not in versionmap[i] and "git" not in versionmap[i]:
                    versionmap[i]["release"] += " (out of date; config issue)"
                    logger.error(
                        "VersionMap: Missing package to be checked",
                        extra={"data": json.dumps(versionmap[i], indent=4)},
                    )
                elif "package" in versionmap[i]:
                    update_from_package(versionmap[i], context)
                elif "git" in versionmap[i]:
                    update_from_git(versionmap[i], context)

            except:
                traceback.print_exc()
                versionmap[i]["release"] += " (out of date; import issue)"
                logger.exception(
                    "{}: Unable to import package/repository to update version".format(
                        label
                    ),
                    extra={"data": json.dumps(versionmap[i], indent=4)},
                )

    # Cleanup of the output
    for i in range(len(versionmap)):

        # Drop the notes by default
        if not include_notes:
            versionmap[i].pop("notes", [])

        if not include_tags:
            versionmap[i].pop("alltags", [])

    # Cache the outputs...
    cached_versionmap[filename] = versionmap

    return timestamp, versionmap



import os
import sys
import json
import re
import hashlib
import pickle
import logging
import operator
import glob
import struct
import traceback
import threading
import time
from datetime import datetime
from collections import Counter, defaultdict
from asgiref.sync import async_to_sync, sync_to_async

from fastapi import FastAPI, BackgroundTasks, Depends

from ..lib import SafeEncoder, slugify
from .log import *

logger = get_logger()

##########################
# Global variables
##########################
stats = {
    "datasets": {}
}
agents = {}
cache = {}

####################
# Helper functions..
#####################
def get_stats():
    return stats

def get_agents():
    return agents

def get_cache():
    return cache

def initialize_stats():

    now = datetime.now().replace(microsecond=0).isoformat()
    stats.update({
        "pid": os.getpid(),
        "cmdline": " ".join(list(sys.argv)),
        "data_root": os.environ['DATA_ROOT'],
        "run_dir": os.environ['RUNDIR'],
        'when': now,
        "datasets": {},
        'query_count': 0,
        'query_success': 0,
        'query_failure': 0,
    })

def get_files(fullpath, exts):

    total = 0
    stats = defaultdict(int)
    files = defaultdict(list)
    for ext in exts:
        pattern = f"{fullpath}/**/*.{ext}"
        inc = glob.iglob(pattern, recursive=True)
        for f in inc:
            if os.path.isfile(f):
                filesize = os.path.getsize(f)
                print(f, filesize)
                if ((ext in ['sqlite', 'db']) or (filesize < (25 * 10**6))): # 10MB
                    files[ext].append(f)
                    stats[ext] += 1
                else:
                    stats[f'{ext}-ignored'] += 1
                stats['total'] += 1
            else:
                stats[f'{ext}-dir'] += 1

    return stats, files


def get_dir_sha1(dir_root, extra=None):

    # Compute a hash of the directory and return its md5 sum

    sha1hash = hashlib.sha1()
    if extra is not None:
        sha1hash.update(bytes(json.dumps(extra), 'utf-8'))

    for dirpath, dirnames, filenames in os.walk(dir_root, topdown=True):

        dirnames.sort(key=os.path.normcase)
        filenames.sort(key=os.path.normcase)

        for filename in filenames:
            filepath = os.path.join(dirpath, filename)

            # 1) filename (good idea)
            sha1hash.update(bytes(os.path.normcase(os.path.relpath(filepath, dir_root)), 'utf-8'))

            # 2) mtime (possibly a bad idea)
            st = os.stat(filepath)
            sha1hash.update(struct.pack('d', st.st_mtime))

            # 3) size (good idea perhaps)
            sha1hash.update(bytes(st.st_size))

    return sha1hash.hexdigest()

#####################################################
# Deprecated...
#####################################################
def read_agent_pickle(name):

    logger.warning("Pickling of agent is deprecated")

    run_root = os.environ['RUNDIR']
    agentname = os.environ.get('AGENTNAME',"default")

    # See if there is a pickle file from the past
    pickledir = os.path.join(run_root, "pickles", agentname)
    try:
        os.makedirs(pickledir)
    except:
        pass
    picklefile = os.path.join(pickledir, f"{name}.pkl")
    if os.path.exists(picklefile):
        try:
            data = pickle.load(open(picklefile, 'rb'))
            logger.info(f"Read agent pickle for {name} from {picklefile}")
            return data
        except:
            logger.exception(f"Failed to read pickle for {name}")

    return None

def write_agent_pickle(name, obj):

    logger.warning("Pickling of agent is deprecated")

    run_root = os.environ['RUNDIR']
    agentname = os.environ.get('AGENTNAME',"default")

    # See if there is a pickle file from the past
    pickledir = os.path.join(run_root, "pickles", agentname)
    try:
        os.makedirs(pickledir)
    except:
        pass
    picklefile = os.path.join(pickledir, f"{name}.pkl")
    try:
        with open(picklefile, 'wb') as fd:
            pickle.dump(obj, fd,
                        protocol=pickle.HIGHEST_PROTOCOL)
            logger.info(f"wrote agent pickle for {name} at {picklefile}")
    except:
        logger.exception(f"Failed to write pickle for {name}")


#####################################################
# Agent details...
#####################################################
def get_llm_cascade():
    llm_cascade = [
        {"id": "economy", "platform": "openai", "model": "gpt-3.5-turbo"},
        {"id": "power", "platform": "openai", "model": "gpt-4"},
    ]
    return llm_cascade

async def get_generic_agent(agentcls,
                            dataset,
                            fullpath,
                            files,
                            metadata,
                            drop_index=False):
    """
    Instantiates the agent.

    Parameters
    ----------
    agentcls: class
              Agent class
    dataset: str
             Name of the dataset
    fullpath: str
              Root directory on the filesystem of dataset
    files: list
           Files within the directory

    Returns
    -------
    status: dict
            Return a diction with agent instance, path of vector index on disk and name of the index
    """

    logger.debug("Using Generic Agent")

    name = os.environ.get('AGENTNAME',"default")
    index_name=slugify(f"{dataset}_index")
    index_path = os.path.expandvars(f"$RUNDIR/$AGENTNAME/index/{dataset}/chromadb_index")
    stats = defaultdict(int)

    agent = agentcls(name=name, cred={})

    try:
        if os.path.exists(index_path) and not drop_index:
            print(f"Loading existing index {name}")
            agent.load_index(store="chroma",
                             persist_directory=index_path,
                             index_name=index_name)
            logger.debug(f"Loaded index: {name}",
                         extra={
                             "data": json.dumps({
                                 'source': 'service',
                                 'user': 'all',
                                 'dataset': dataset,
                                 'stats': agent.get_index_stats()
                             }, indent=4, cls=SafeEncoder)
                         })

            return {
                'agent': agent,
                'index_path': index_path,
                'index_name': index_name
            }

    except:
        logger.exception(f"Failed to load index: {name}")

    print(f"Building new index {name}")
    logger.debug(f"Building a new index",
                 extra={
                     "data": json.dumps({
                         'source': 'service',
                         'user': 'all',
                         'dataset': dataset,
                         'data': index_path
                     },  indent=4, cls=SafeEncoder)
                 })

    first = True
    for ext, extfiles in files.items():

        params = {}
        if ext == "pdf":
            params['pdfloader'] = "pymupdf"

        for f in extfiles:

            # Load data...
            data = None
            try:
                if ext == "pdf":
                    data = agent.load_data(source='pdf',
                                           content=f,
                                           params={
                                               "pdfloader": "pymupdf"
                                           })
                    stats[f'{ext}-mupdf'] +=1
                elif ext in ["doc", "docx"]:
                    data = agent.load_data(source='docx', content=f,)
                    stats[f'{ext}-default'] +=1
                else:
                    data = agent.load_data(source='dir',
                                           content=os.path.dirname(f),
                                           params={
                                               "glob": os.path.basename(f)
                                           })
                    stats[f'{ext}-dirload'] +=1
            except:
                traceback.print_exc()
                stats[f'{ext}-error'] += 1

            if data is None:
                continue

            if first:
                print(f"Creating new index: {name}")
                agent.create_add_index(data=data,
                                       store="chroma",
                                       persist_directory=index_path,
                                       index_name=index_name)
                first = False
            else:
                agent.add_to_index(data=data)

    print(f"Loaded agent {name}")
    logger.debug(f"Loaded agent",
                 extra={
                     "data": json.dumps({
                         'source': 'service',
                         'user': 'all',
                         'dataset': dataset,
                         'stats': agent.get_index_stats()
                     }, indent=4, cls=SafeEncoder)
                 })

    return {
        'agent': agent,
        'index_path': index_path,
        'index_name': index_name
    }

async def get_agent_details(namespace,
                            username,
                            subdir,
                            exts,
                            get_task_specific_agent,
                            force=False,
                            extra=None):
    """
    Check agent cache and return if it exists. Else instantiate an agent. Updates the statisics as well

    Parameters
    ----------
    namespace: str
               Enable instantiating the agent multiple times using a namespace string
    subdir: str
            Subdirectory within the DATA_ROOT environment variable
    username: str
              Name of the user. Each user has a separate instance
    exts: list
           List of extensions
    get_task_specific_agent: callable
           Callback to instantiate an agent

    Returns
    -------
    status: dict
            Return a diction with agent instance, path of vector index on disk and name of the index
    """

    label = f"{username}_{subdir}"
    now = datetime.now().replace(microsecond=0).isoformat()

    run_root = os.environ['RUNDIR']
    dir_root = os.environ['DATA_ROOT']
    fullpath = os.path.join(dir_root, namespace, subdir)
    cred = {}

    # Add a check here to see if something has changed...
    reindex = os.path.join(fullpath, '.reindex')
    print("Reindex check", reindex)
    if os.path.exists(reindex):
        print("Document set changed. So indexing again")
        force = True

    if ((label in agents) and (not force)):
        logger.debug(f"[{namespace}] Returning existing instance",
                     extra={
                         'source': 'service',
                         'user': username,
                         'dataset': subdir
                     })
        return agents[label]

    try:

        if ((label in stats['datasets']) and (not force)):
            logger.debug(f"Return existing agent",
                         extra={
                             'source': 'service',
                             'user': username,
                             'dataset': subdir,
                             'data': json.dumps(stats['datasets'][label],
                                                indent=4,
                                                cls=SafeEncoder)
                         })
        else:
            logger.debug(f"Creating agent (Force: {force})",
                         extra={
                             'source': 'service',
                             'user': username,
                             'dataset': subdir,
                             'data': fullpath
                         })

        if not os.path.isdir(fullpath):
            logger.debug(f"Could not find data",
                         extra={
                             'source': 'service',
                             'user': username,
                             'dataset': subdir,
                             'data': fullpath
                         })
            return None

        # Compute the sha1
        sha1 = get_dir_sha1(fullpath, extra)

        stats['datasets'][label] = {
            'sha1': sha1,
            'loaded': now,
            'username': username,
            'subdir': subdir,
            'fullpath': fullpath,
            'agent_created': False,
            'agent_status': "initialization",
            "query_count": 0,
            "query_success": 0,
            "query_failure": 0,
        }

        filestats, files = get_files(fullpath, exts)

        logger.debug(f"Dataset details",
                     extra={
                         'source': 'service',
                         'user': username,
                         'dataset': subdir,
                         'data': json.dumps({
                             "filestats": filestats,
                             "files": files
                             }, indent=4)
                         })

        if filestats['total'] == 0:
            logger.warning(f"No data found",
                           extra={
                               'source': 'service',
                               'user': username,
                               'dataset': subdir,
                        })

            stats['datasets'][label]['agent_status'] = f"Error! Files not found"
            return None

        # Include the metadata
        metadatapath = os.path.join(fullpath, 'metadata.json')
        print("metadatapath", metadatapath)

        metadata = {}
        if os.path.exists(metadatapath):
            try:
                metadata = json.load(open(metadatapath))
            except:
                logger.exception(f"Failed to read metadata",
                                 extra={
                                     'source': 'service',
                                     'user': username,
                                     'dataset': subdir,
                                     'data': metadatapath
                                 })


        stats['datasets'][label].update({
            'filestats': filestats,
            "files": files,
            "metadata": metadata
        })


        agentdetails = await get_task_specific_agent(namespace,
                                                     username,
                                                     subdir,
                                                     fullpath,
                                                     files,
                                                     metadata,
                                                     drop_index=force)

        if agentdetails is None:
            raise Exception("Task specific agent returned invalid details")


        stats['datasets'][label].update({
            'agent_created': True,
            "agent_status": "Created"
        })

        agents[label] = {
            'sha1': sha1,
            'created': now,
            "metadata": metadata
        }

        agents[label].update(agentdetails)

        stats['datasets'][label]['agent_status'] = f"Ready!"
        logger.debug(f"Agent is ready!",
                       extra={
                           "data": {
                               'source': 'service',
                               'user': username,
                               'dataset': subdir,
                               'data': metadatapath
                           }
                       })

        if os.path.exists(reindex):
            os.remove(reindex)

    except Exception as e:
        logger.exception(f"Failed to build agent",
                       extra={
                           "data": {
                               'source': 'service',
                               'user': username,
                               'dataset': subdir,
                           }
                       })

        stats['datasets'][label]['agent_created'] = False
        stats['datasets'][label]['agent_status'] = f"Error! {e}"
        return None

    return agents[label]


async def check_agent_details(username, subdir):
    label = f"{username}_{subdir}"
    return label in agents

###########################################################
# Build tasks..
###########################################################

buildtasks = {}
def get_buildtask_details(label):
    return buildtasks.get(label, None)

def add_buildtask(label, params):

    if label not in buildtasks:
        condition.acquire()
        buildtasks[label] = params
        condition.release()
        msg = "Started indexing documents in the background and creating an agent"
    else:
        status = buildtasks[label]['status']
        if status == "tobuild":
            msg = 'Index is being built. Wait'
        elif status == "failure":
            msg = 'Index could not be built. See log'
        elif status == "success":
            msg = 'Index has been built. Use force is rebuild'
        else:
            msg = 'Status of index is unknown. Could be an internal error. Check log'

    return msg

condition = threading.Condition()
class BuildAgentThread(threading.Thread):

    #Background job that is continuously building the indexes
    #required

    def __init__(self, agentfunc, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.agentfunc = agentfunc

    def run(self, *args,**kwargs):
        global buildtasks

        agentfunc = self.agentfunc

        logger.debug(f"Index builder thread Start {id(buildtasks)}")
        while True:
            for label in list(buildtasks.keys()):
                params = buildtasks[label]
                status = params.get('status','unknown')
                force = params.get('force', False)
                if ((status == "tobuild") or (force)):
                    logger.debug(f"Building index for {label} (force={force})")
                    try:
                        details = async_to_sync(get_agent_details)\
                            (namespace=params['namespace'],
                             username=params['user'],
                             subdir=params['dataset'],
                             exts=params['exts'],
                             get_task_specific_agent=agentfunc,
                             force=force)

                        status = "failure" if details is None else "success"
                        message = "Regular execution"
                        logger.debug(f"Index Build: {status}",
                                     extra={
                                         "source": "service",
                                         'user': params['user'],
                                         "dataset": params['dataset'],
                                         "data": json.dumps(params, indent=4, cls=SafeEncoder)
                                     })
                    except Exception as e:
                        logger.exception(f"Index Build: failure",
                                         extra={
                                             "source": "service",
                                             'user': params['user'],
                                             "dataset": params['dataset'],
                                             "data": json.dumps(params, indent=4, cls=SafeEncoder)
                                         })
                        status = "failure"
                        message = f"Exception {str(e)}"

                    condition.acquire()
                    params['status'] = status
                    params['message'] = message
                    condition.release()

            time.sleep(0.1)

#####################################################
# Helper functions
#####################################################
def add_show_cache(app):

    @app.get("/cache")
    def show_cache():
        """
        Summarize the cache
        """

        cache = get_cache()
        summary = defaultdict(int)
        for k, v in cache.items():
            summary[v['status']] += 1

        return summary

def add_show_health(app):

    @app.get("/health")
    def show_health():
        """
        Return the stats
        """
        logger.info("Returning stats",
                    extra={
                        "source": "service"
                    })
        return stats

def add_build_index(app, namespace_default, exts):

    @app.get("/index/build")
    async def build_index(user: str,
                          dataset: str,
                          background_tasks: BackgroundTasks,
                          namespace=namespace_default,
                          force=False
    ):

        label = f"{user}_{dataset}"

        params = {
            "status": "tobuild",
            "user": user,
            "dataset": dataset,
            "namespace": namespace,
            "exts": exts
        }

        # => Insert it into the build tasks if not already indexed...
        msg = add_buildtask(label, params)

        logger.debug(msg,
                     extra={
                         "source": "service",
                         "user": user,
                         "dataset": dataset,
                     })
        return {
            'status': 'success',
            'message': msg
        }

def add_check_index(app, namespace_default):

    @app.get("/index/check")
    async def check_index(user: str,
                          dataset: str,
                          namespace=namespace_default
    ):
        """
        Check build tasks
        """

        label = f"{user}_{dataset}"
        try:

            params = buildtasks.get(label,{})
            status = params.get('status', 'unknown')
            if len(params) == 0:
                msg = "Does not exist"
            elif status == "success":
                msg = 'Exists'
            else:
                msg = "Does not exist"
            logger.debug(f"Index check: {status}",
                         extra={
                             "source": "service",
                             "user": user,
                             "dataset": dataset,
                        })

            return {
                'status': 'success',
                'message': msg
            }
        except Exception as e:
            logger.exception(f"Index check: Failed.",
                             extra={
                                 "source": "service",
                                 "user": user,
                                 "dataset": dataset,
                             })
            return {
                'status': 'failure',
                'message': f"Internal error: {str(e)}"
            }



def query_update_result(request_id, result):

    # First get the params
    cache = get_cache()
    value = cache[request_id]

    for k, v in result.items():
        value[k] = result[k]

    status = result.get('status', 'unknown')
    logger.debug(f"Updated Result: {status}",
                extra={
                    "source": "service",
                    "request_id": request_id,
                    "user": value.get("user", "unknown"),
                    "dataset": value.get("dataset", "unknown"),
                    'data': json.dumps(result, indent=4, cls=SafeEncoder)
                })

###############################################
#
###############################################
def add_policy(app, policyspec):

    @app.get("/policy")
    def show_policy():
        """
        Return the stats
        """

        logger.info("Returning policy",
                    extra={
                        "source": "service"
                    })
        return {
            "status": "success",
            "data": policyspec
        }



import os
import re
import json
import traceback
import logging
import seaborn as sns
import matplotlib.pyplot as plt
from io import BytesIO, StringIO
import base64

from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from django.http import QueryDict
from django.shortcuts import render
from django.http import HttpResponse, JsonResponse
from django.conf import settings
from django.db import models
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_http_methods
from django.http import HttpResponseRedirect, Http404
from django.contrib import messages
from django.urls import reverse, resolve
from django.conf.urls import url, include

import s3fs
import gcsfs

from users.decorators import log_activity
from dashboard.lib import log_app_activity

from enrichsdk.utils import SafeEncoder
from enrichapp.spec import validate_spec as default_validate_spec, fill_action_gaps
from enrichsdk.lib import get_credentials_by_name
from enrichsdk.utils import SafeEncoder
from enrichsdk.datasets import Doodle
from enrichsdk.lib import get_default_from_email
from enrichsdk.lib.integration import send_html_email

logger= logging.getLogger('app')


###############################################
# Main class
###############################################
class AppBase():
    """
    A shareable view across apps
    """

    def __init__(self):
        self.name = "singlepageapp"
        self.verbose_name = "Single Page App"
        self.category = "singlepageapp"
        self.templates = {
            'index': f'enrichapp/singlepageapp/index.html',
            'helper': f'enrichapp/singlepageapp/helper.html',
            'workflow_index': f'enrichapp/policyapp/workflow_index.html',
        }

    def __str__(self):
        return f"[{self.name}] {self.verbose_name}"

    def get_readme(self):
        return """This is a generic single page app"""

    def get_template(self, spec, name):
        return spec.get('templates',{}).get(name, self.templates.get(name, "Unable to find template"))

    def get_model(self, spec, name):

        from enrichapp.dashboard.globalapp import lib as globallib

        if (('models' in spec) and (name in spec['models'])):
            return spec['models'][name]

        return globallib.get_model(name)

    def get_form(self, spec, name):

        from enrichapp.dashboard.globalapp import lib as globallib

        if (('forms' in spec) and (name in spec['forms'])):
            return spec['forms'][name]

        return globallib.get_form(name)

    def validate_spec(self, spec):

        fill_action_gaps(spec, context={ })

        errors = default_validate_spec(spec)
        if len(errors) > 0:
            return len(errors) == 0, errors

        return len(errors) == 0, errors

    def check_prerequisites(self, request, spec):

        valid, errors = self.validate_spec(spec)
        if not valid:
            error = "Invalid specification"
            messages.error(request, error)
            msg = "Errors: " + json.dumps(errors, indent=4, cls=SafeEncoder)
            msg += "Spec:\n" + json.dumps(spec,  indent=4, cls=SafeEncoder)

            logger.error(error,
                         extra={
                             'transform': 'Dashboard',
                             'data': msg
                         })
            return valid, HttpResponseRedirect(reverse('dashboard:application_index'))

        return valid, None

    def get_urlpatterns(self):
        urlpatterns = [
            url('^workflows/action[/]?$', self.workflow_action, name="workflow_action"),
            url('^[/]?$', self.index, name="index"),
        ]

        return urlpatterns

    ################################################################
    # Workflow Management
    ################################################################
    def get_supported_actions(self, request, spec):
        return ['Email']

    def take_action(self, request, spec, data, target):

        r = resolve(request.path)

        detailurl = request.build_absolute_uri(reverse(r.namespace + ":index")) + "?" + request.GET.urlencode()

        try:
            if target.nature == "Email":

                """{
                    "emails": [
                       "pingali@scribbledata.io"
                     ]
                }"""

                policy = target.policy
                if (('emails' not in policy) and (not isinstance(target.policy['emails'], list))):
                    return JsonResponse({
                        'status': 'failure',
                        'message': "Email service not configured correct. Please contact the admin"
                    })

                user = request.user
                fullname = f"{user.first_name} {user.last_name}".strip()
                if len(fullname) == 0:
                    fullname = user.username

                sender = fullname + "<" + get_default_from_email() +">"
                receivers = policy['emails']
                subject = f"{spec['name']} - Sharing Observation"
                content = f"<p>{data['message']}</p>"
                content += f"<p>You can find the full details at <a href='{detailurl}'>{spec['name']}</a> app."
                if (('data' in data) and (len(data) > 0)):
                    content += "Here is quick context:</p>"
                    content += "<p>" + data['data'] + "</p>"

                msg = MIMEMultipart()
                msg.attach(MIMEText(content, "html"))
                send_html_email(content=content,
                                sender=sender,
                                receivers=receivers,
                                subject=subject)

                return JsonResponse({
                    'status': 'success',
                    'message': "Email sent successfully"
                })

            return JsonResponse({
                'status': "failure",
                "message": f"Workflow service not configured correctly. Please contact admin"
            })

        except Exception as e:
            traceback.print_exc()
            logger.exception("Unable to send email")
            return JsonResponse({
                'status': "failure",
                "message": f"{target.nature} message action could not be taken. {str(e)}"
            })

    @log_activity("workflow_action", nature="application")
    def workflow_action(self, request, spec):
        """
        Show the workflows for the given data
        """

        from dashboard.models import AppActionTarget

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']

        data = request.POST.dict()

        try:
            try:
                target = AppActionTarget.objects.get(pk=data['actionid'])
            except:
                raise Exception("Internal error. Missing or invalid action id")

            # messages.success(request, "Processed the post")
            status = "success"

            # Make this happen...
            status = self.take_action(request, spec, data, target)

        except Exception as e:
            return JsonResponse({
                'status': 'failure',
                'message': f"Failed to trigger action. {str(e)}"
            })

        return status


    ################################################################
    # Data Access
    ################################################################
    def get_fs_handle(self, cred):
        """
        Get s3/gcs filesystem handle..
        """
        cred = get_credentials_by_name(cred)
        nature = cred['nature']
        if nature not in ['s3', 'gcs']:
            raise Exception(f"Unknown credentials: {nature}")

        if nature == 's3':
            config_kwargs={
                'signature_version': 's3v4'
            }

            if 'region' in cred:
                config_kwargs['region_name'] = cred['region']

            fshandle = s3fs.S3FileSystem(
                key    = cred['access_key'],
                secret = cred['secret_key'],
                config_kwargs=config_kwargs,
                use_listings_cache=False
            )
        else:
            fshandle = gcsfs.GCSFileSystem(
                token=cred['keyfile']
            )

        return fshandle

    def get_data(self, request, spec):

        try:
            data = {
                "test": "This is test data"
            }

        except:
            logger.exception(f"Error in accessing {spec['name']} results")

        return data

    @log_activity("app_index", nature="application")
    def index(self, request, spec):

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']

        template = self.get_template(spec, 'index')
        helper = self.get_template(spec, 'helper')

        try:
            data = self.get_data(request, spec)
        except:
            logger.exception("Unable to read data")
            messages.error(request, "Unable to read data. See log")
            return HttpResponseRedirect(reverse('dashboard:application_index'))

        return render(request,
                      template,
                      {
                          'app': self,
                          'spec': spec,
                          'usecase': usecase,
                          'basenamespace': r.namespace,
                          'data': data,
                          'helper': helper
                      })


    def get_router(self):

        from ninja import Router, Query, Schema

        router = Router()

        # Nothing to do...
        return router



from django import forms
from .models import *
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout

class AppPolicyFormBase(forms.ModelForm):
    notes = forms.CharField( widget=forms.Textarea, required=False)
    class Meta:
        model = AppPolicyBase
        exclude = ('appname', 'namespace', 'created_by', 'modified_by')
        widgets = {
          'config': forms.Textarea(attrs={'rows':150, 'cols':200}),
        }



import os
import json
import time
from collections import defaultdict

from langchain import LLMChain
from langchain import PromptTemplate
from langchain.chains.question_answering import load_qa_chain
from langchain.memory import ConversationTokenBufferMemory, ConversationKGMemory
from langchain.callbacks import get_openai_callback

from . import agent_events
from llmsdk.agents.basellm import BaseLLMAgent


__all__ = ['LLMIndexQuerier']


class LLMIndexQuerier(BaseLLMAgent):
    """
    Class to do querying using LLMs
    Query can be run against a specified set of documents that act
    as context to constrain the answers
    or against all the stored knowledge of the LLM model
    """

    def __init__(self,
                 name,
                 cred={},
                 platform="openai",
                 searchapi="serpapi",
                 statestore="redis",
                 memory_size=1000):
        """
        init the LLM query agent
        name: name of the agent
        cred: credentials object
        platform: name of the platform backend to use
                default to OpenAI GPT model for now, and Azure as well
                will be extended in the future to suuport other models
        memory_size: how many tokens of memory to use when chatting with the LLM
        """

        start_time = time.time()

        # init the base class
        super().__init__(name=name,
                         cred=cred,
                         platform=platform,
                         agent_type="qna",
                         searchapi=searchapi,
                         statestore=statestore,
                         memory_size=memory_size)

        # defaults
        self.chunk_size = 1000
        self.chunk_overlap = 200
        self.index = None
        self.latest_context = []
        self.context_topK = 1
        self.current_kg = []
        self.metadata = {}
        self.vdb_client = None
        self.index_name = None
        self.index_store = None
        self.state_store = statestore
        self.docs = {}

        # LLM params
        self.platform = platform
        self.searchapi = searchapi
        self.chaintype = "stuff"
        self.memory_size = memory_size

        # init the memories for the llm
        conv_memory = ConversationTokenBufferMemory(llm=self.llm,
                                                    max_token_limit=self.memory_size,
                                                    memory_key="chat_history",
                                                    input_key="input",
                                                    return_messages=True)
        kg_memory = ConversationKGMemory(llm=self.llm,
                                         return_messages=True)
        self.memory = conv_memory
        self.kg = kg_memory

        # init the QnA chain for internal queries
        prompt = self._get_query_prompt_internal()
        self.llm_chain_int = load_qa_chain(llm=self.llm,
                                           chain_type=self.chaintype,
                                           memory=self.memory,
                                           prompt=prompt)
        # init the chain for external queries
        prompt = self._get_query_prompt_external()
        self.llm_chain_ext = LLMChain(llm=self.llm,
                                      prompt=prompt)
        # init the chain for kwords
        prompt = self._get_query_prompt_kwords()
        self.llm_chain_kw = LLMChain(llm=self.llm,
                                      prompt=prompt)
        # init the chain for suggestions
        prompt = self._get_query_prompt_suggest()
        self.llm_chain_sug = LLMChain(llm=self.llm,
                                      memory=self.memory,
                                      prompt=prompt)
        # init the agent for searches
        self.llm_agent_srch, self.searchengine = self._load_search_agent(cred=self.cred,
                                                      searchapi=self.searchapi,
                                                      llm=self.llm)
        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "type": self.agent_type,
                "id": self.agent_id,
                "platform": self.platform,
                "searchapi": self.searchapi,
                "statestore": self.state_store,
                "memory_size": self.memory_size,
                "chaintype": self.chaintype,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)

    ## helper functions

    def _get_query_prompt_internal(self):
        """
        generate a prompt for running a query in internal mode
        """
        template = """You are a helpful and diligent chatbot having a conversation with a human.
Use the following pieces of context to answer the question at the end.
If you don't know the answer, just say 'I am not sure', don't try to make up an answer.
If the question is framed in active voice, first convert it to passive voice and use that to answer the question.

        {context}

        {chat_history}
        Human: {input}
        Chatbot:"""

        prompt = PromptTemplate(
            input_variables=["chat_history", "input", "context"],
            template=template
        )

        return prompt

    def _get_query_prompt_external(self):
        """
        generate a prompt for running a query in external mode
        """
        template = """In the context of the following terms: {context},

        {input}"""

        prompt = PromptTemplate(
            input_variables=["context", "input"],
            template=template
        )

        return prompt

    def _get_query_prompt_suggest(self):
        """
        generate a prompt for running a query to suggest other ways of asking a question
        given some chat history
        """
        template = """You are a helpful and diligent chatbot having a conversation with a human.
        Your task is to only suggest some alternate ways to pose the last question based on the context in the chat history.
        Format your response as a json list only.

        {chat_history}
        Human: {input}
        Chatbot:"""

        prompt = PromptTemplate(
            input_variables=["chat_history", "input"],
            template=template
        )

        return prompt

    def _get_query_prompt_kwords(self, context=""):
        """
        generate a prompt for extracting keywords from a paragraph
        """
        template = """You are the chief of staff to a busy executive. You will be given a paragraph of text
        and must identify a few (maximum of 5) key phrases from the paragraph. Do not summarize what you find,
        respond with the key phrases exactly. Structure your response as a json list exactly.
        DO NOT RESPOND WITH ANY KEY PHRASES FROM CONTENT ABOVE THIS LINE.

        Here is the paragraph:
        {context}

        The key phrases are:
        """

        prompt = PromptTemplate(
            input_variables=["context"],
            template=template,
        )

        return prompt

    def _list_to_nested_dict(self, l):
        """
        take a list and return a nested dict
        used only for formating the KG entities
        """
        # loop through them to construct the dict
        rd = defaultdict(lambda: defaultdict(list))
        for e in l:
            # get constiuents
            subject = e[0]
            predicate = e[1]
            object_ = e[2]

            # format the output
            rd[subject][predicate].append(object_)

        def default_to_regular(d):
            if isinstance(d, defaultdict):
                d = {k: default_to_regular(v) for k, v in d.items()}
            return d

        return default_to_regular(rd)

    ## interfaces

    def run_query_internal(self, query):
        """
        run a query using llm on an internal docset indexed in index
        this is useful when looking for answers using a private source of data
        """
        # get the similar docs
        docs = self.get_similar_docs(query)

        if docs is None:
            return {
                "question": query,
                "answer": "Could not find document chunks to process"
            }

        # setup the QnA chain object
        response = self.llm_chain_int({"input_documents":docs, "input":query},
                                    return_only_outputs=False)

        # run the query against the similar docs
        result = {
            "question": query,
            "answer": response.get('output_text', self._err_msg('field')).strip(),
            "sources": [{"content": d.page_content, "metadata": d.metadata, "distance": d.metadata.pop('distance')} for d in docs],
        }

        # check if suggest call is needed
        if ('output_text' not in response) or ("i am not sure" in result['answer'].lower()):
            response = self.run_query_suggest(query)
            result['suggest'] = response['suggest']
            # we don't have a usable answer, so no need for sources
            result['sources'] = []

        return result

    def run_query_external(self, query):
        """
        run a query using llm
        this is useful when looking for answers that generic llm can provide
        """
        # augment the query with some context to guide the LLM
        context = ", ".join(self.latest_context)
        result = self.llm_chain_ext({"context": context, "input":query},
                                    return_only_outputs=True)
        result = {
            "question": query,
            "answer": result,
            "sources": [{"source": f"llm-{self.platform}"}]
        }

        return result

    def run_query_search(self, query):
        """
        run a query using the search agent
        this is useful when looking for answers using a search engine
        """
        def extract_content_sources(sourcedata):
            docs = sourcedata.get('organic_results')
            if not docs:
                return None
            sources = [{"content": d.get('snippet', ""), "source": d.get('link')} for d in docs]
            return sources

        # modify the query using context history
        context = ", ".join(self.latest_context)
        query_mod = f"In the context of {context}, {query}"

        # get the human-readable result
        result = self.llm_agent_srch.run(input=query_mod)

        # get the sources
        sourcedata = self.searchengine.results(query_mod)
        sources = extract_content_sources(sourcedata)
        if not sources:
            sources = [{"content": "", "source": f"search-{self.searchapi}"}]

        # construct result
        result = {
            "question": query,
            "answer": result,
            "suggest": list(set([q.get('question', '') for q in sourcedata.get('related_questions', [])])),
            "sources": sources
        }

        return result

    def run_query_suggest(self, query):
        """
        run a query using llm to suggest other ways of asking the query,
        in the context of the chat history
        """
        # augment the query with some context to guide the LLM
        response = self.llm_chain_sug({"input":query},
                                      return_only_outputs=True)
        result = response.get('text', response).strip()
        try:
            # we asked the LLM to give us json
            suggest = json.loads(result)
            # in case the LLM gave us a list of dicts instead of list of strs
            suggest = [list(s.values())[0] if isinstance(s, dict) else s for s in suggest]
        except:
            suggest = []

        result = {
            "question": query,
            "answer": result,
            "suggest": suggest,
            "sources": [{"source": f"llm-{self.platform}"}]
        }

        return result

    def run_query_kwords(self, context=""):
        """
        run a query using llm on an internal docset indexed in index
        this is useful when looking for answers that generic llm can provide
        """
        result = self.llm_chain_kw({"context": context},
                                    return_only_outputs=True)
        result = result.get('text', result)
        # a few tries to extract the response
        # sometimes, the LLM messes up
        try:
            result = json.loads(result)
        except:
            try:
                result = json.loads(f"[{result.split('[')[-1]}")
            except:
                try:
                    result = result.split("\n")
                    result = [r.replace("- ", "") for r in result]
                except:
                    pass

        # force a list
        result = [] if not isinstance(result, list) else result

        return result

    def extract_add_kg_entities(self, answer):
        """
        extract all KG entities, format as {entity: relation}
        and add to the current set of tracked KG entities
        """
        # get the KG entities for this answer
        kg_entities = self.kg.get_knowledge_triplets(answer)
        # format the entities as we need them
        kge = [(e.subject, e.predicate, e.object_) for e in kg_entities]

        ## -- TO-DO --
        # at this point
        # we will have to do some post-processing on the KG entities
        # one option is to look at useful relations
        # defined by the predicate
        # e.g. {"is defined as", "shall mean", "is"} -> "="
        ## -- END TO-DO --

        # add to running list of all KG entities
        self.current_kg += kge

        # format for output
        kge = self._list_to_nested_dict(kge)

        return kge

    def get_kg_entities(self):
        """
        return all the KG entities as a dict
        {entity: (relation, object)}
        """
        # get the current list of KG entities
        kg_entities = self._list_to_nested_dict(self.current_kg)

        return kg_entities


    def query(self, query, mode="internal", policy={}):
        """
        run a query on an index using an llm chain object
        query: query string
        mode: 'internal' for querying over docset,
              'external' for general query,
              'suggest' for asking the LLM for alternate ways to pose the question
        policy: any extra params needed by the agent
        """

        start_time = time.time()

        # check to see if we need to reset agent memory
        reset_state = policy.get("reset_state", False)
        if reset_state:
            # clear the agent state
            self.clear_agent_state()
        else:
            # set the state of the agent
            # this happens only if we are not doing a reset-state
            self.set_agent_state(statekey=self.statekey)

        method = getattr(self, f"run_query_{mode}", None)
        if method is None:
            raise Exception(f"Unsupported mode: {mode}")

        if self.platform in ['openai', 'azure']:
            with get_openai_callback() as cb:
                result = method(query)
            stats = {
                "total_tokens": cb.total_tokens,
                "prompt_tokens": cb.prompt_tokens,
                "completion_tokens": cb.completion_tokens,
                "total_cost": round(cb.total_cost, 4)
            }
        else:
            result = method(query)
            stats = {}

        if result:

            answer = result['answer']

            if answer:
                # add keywords identified to the result
                result['keywords'] = self.run_query_kwords(context=answer)

                # store the latest context
                # this is useful to guide the external agent
                # since it is memory-less
                # only store the top-n keywords
                # storing more will make the LLM overfit responses
                self.latest_context = result['keywords'][0:self.context_topK]

                # add KG elements to the result
                result['kg'] = self.extract_add_kg_entities(answer)

        # store the agent's state
        self.store_agent_state()

        # log the event
        params = {
            "query": query,
            "mode": mode,
            "policy": policy,
            "result": result.copy() if result is not None else None,
            "stats": stats
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        # add the event to the result
        if result:
            result['metadata'] = {
                "timestamp": event['timestamp'],
                "duration": event['duration'],
                "stats": stats
            }

        return result


if __name__ == "__main__":

    dirpath = '...'
    index_name = "bf_index"
    persist_directory = "chromadb123"

    queries = [
        "what are the top heavy rules?",
        "which mortality table to use for top heavy rules?",
        "tell me about plan termination",
        "tell me about the retirement plan termination",
        "are there any death benefits?",
        "explain employer contributions",
        "are there any participant contributions needed?",
        "explain participant contributions",
        "who is optimus prime?",
        "tell me about the benefits committee",
        "what is the expansion of UCN?",
        "what is the expansion of UCN? if the answer is not found in the documents, say so",
        "what is the definition of a plan participant?",
        "what sections mention the retirement age?",
        "what is the retirement age?",
        "are there clauses for the retirement age?",
        "how is the pension calculated?"
    ]
    query = queries[-1]

    # setup the agent and the index
    agent = LLMIndexQuerier(name="agent_bf", platform="azure")
    # point it to the data
    # data = agent.load_data(source="dir",
    #                        content=dirpath,
    #                        params={"glob":"**/*.*"})
    data = agent.load_data(content=dirpath)

    # create the index
    agent.create_add_index(data=data,
                           store="chroma",
                           persist_directory=persist_directory,
                           index_name=index_name)

    # start asking questions
    query = "what is the deadline for prelim quotes?"

    # internal mode
    # run query against the index
    result = agent.query(query, mode="internal")
    print ("Internal:")
    print (json.dumps(result, indent=4))

    # # external mode
    # # run query against world knowledge
    # result = agent.query(query, mode="search")
    # query = "tell me about that table"
    # print ("Search:")
    # print (json.dumps(result, indent=4))
    #
    # # internal mode
    # # run query against the index
    # query = "who is optimus prime?"
    # result = agent.query(query, mode="internal")
    # print ("Internal with suggestions:")
    # print (json.dumps(result, indent=4))
    #
    # # suggest mode
    # # run query against the index
    # # clear the memory of the agent to simulate a new session of QnA
    # policy = {
    #     "reset_state": True
    # }
    # query = "What is the situs state for this plan?"
    # result = agent.query(query, mode="suggest", policy=policy)
    # print ("Suggestions:")
    # print (json.dumps(result, indent=4))
    #
    # # get the current set of KG entities
    # kg_entities = agent.get_kg_entities()
    # print ("KG:")
    # print(json.dumps(kg_entities, indent=4))

    # get the current metadata
    metadata = agent.get_metadata()
    print ("Metadata:")
    print(json.dumps(metadata, indent=4))



import json
import os
from collections import defaultdict
from datetime import datetime
from uuid import uuid4

from fastapi import APIRouter
from starlette.background import BackgroundTasks
from starlette.concurrency import run_in_threadpool

from llmsdk.lib import SafeEncoder
from llmsdk.services.lib import get_cache, initialize_stats, query_update_result, stats
from llmsdk.services.log import get_logger

logger = get_logger()


class BaseGPT():
    def __init__(self):
        """
        Base GPT services class for rolling out the microservices. Provides common routing methods

        """
        initialize_stats()
        self.name = "BaseGPT"
        self.agentname = "basegptname"
        self.platform = "openai"
        self.namespace = "basegpt"

        self.run_dir = stats.get('run_dir', self.agentname)
        self.index_name = "acme_gpt_index"
        self.persist_directory = os.path.join(self.run_dir, "index")
        self.agent = None

        self.router = APIRouter()
        self.router.add_api_route("/up", self.up, methods=["GET"])
        self.router.add_api_route("/health", self.health, methods=["GET"])
        self.router.add_api_route("/cache", self.cache_health, methods=["GET"])
        self.router.add_api_route("/qna/status", self.qna_status, methods=["GET"])
        self.router.add_api_route("/qna", self.qna, methods=["POST"])
        self.router.add_event_handler("startup", self.app_startup)

    def up(self) -> str:
        """
        Checks if service is running
        :return: Service status
        """
        return f"Service {self.agentname} is working!"

    def health(self) -> dict:
        """
        Return usage statistics

        Returns
        -------
        stats: dict
               Usage statistics

        """
        logger.info("Returning stats",
                    extra={
                        'data': json.dumps(stats, indent=4)
                    })
        return stats

    def cache_health(self):
        """
        Summarize the content of the cache
        """

        cache = get_cache()
        summary = defaultdict(int)
        for k, v in cache.items():
            summary[v['status']] += 1

        return summary

    def get_customer_profile_spec(self) -> dict:
        """
        Profile spec for each customer to be used
        :return:
        """
        pass

    async def get_llm_agent(self) -> dict:
        """
        Get the LLM agent configuration class
        :return: agent configuration
        """
        pass

    def startup_extra_steps(self):
        """
        Run additional startup steps when app is initiated
        :return:
        """
        pass

    async def qna_run_execute(self, query):
        """
        Perform operation using agent, override for different type of operation
        :param query: Query to be searched
        :return:
        """
        spec = self.get_customer_profile_spec()
        return await run_in_threadpool(lambda: self.agent.query(query, spec))

    async def qna_run(self, request_id):
        """
        Execute query against the agend
        :param request_id: Request ID of the request
        :return:
        """
        cache = get_cache()

        if request_id not in cache:
            logger.error(f"Failure",
                         extra={
                             'request_id': "invalid",
                             "source": "service",
                         })
            cache[request_id] = {
                'status': "failure",
                "message": f"Invalid request id"
            }
            return

        # First get the params
        value = cache[request_id]

        try:

            params = value['params']
            user = params['user']
            dataset = params['dataset']
            query = params['query']

            stats['query_count'] += 1

            label = f"{user}_{dataset}"

            # First get the agent...
            logger.debug(f"Getting agent",
                         extra={
                             "source": "service",
                             "user": user,
                             "dataset": dataset,
                             "request_id": request_id,
                             'data': json.dumps(value, indent=4, cls=SafeEncoder)
                         })

            stats['datasets'][label] = {
                'loaded': datetime.now().replace(microsecond=0).isoformat(),
                'username': user,
                'agent_created': True,
                'agent_status': "Created",
                "query_count": 0,
                "query_success": 0,
                "query_failure": 0,
            }

            spec = self.get_customer_profile_spec()

            # Now run the query
            result = await self.qna_run_execute(query)

            if result.get("success"):
                key = result.get("key")
                logger.debug(f"Query: {query}\nKEY: {key}\n")
                json_result = json.loads(json.dumps(result, indent=4, cls=SafeEncoder))

                stats['query_success'] += 1
                stats['datasets'][label]['query_count'] += 1
                stats['datasets'][label]['query_success'] += 1
                query_update_result(request_id, {
                    "status": "success",
                    "result": json_result,
                })
            else:
                suggestions = result.get("suggestions")
                logger.debug(f"Query: {query}\nSuggestions: {suggestions}\n")
                json_result = json.loads(json.dumps(result, indent=4, cls=SafeEncoder))
                stats['datasets'][label]['query_count'] += 1
                stats['datasets'][label]['query_failure'] += 1
                query_update_result(request_id, {
                    "status": "success",
                    "result": json_result,
                })

        except Exception as e:
            stats['query_failure'] += 1
            stats['datasets'][label]['query_count'] += 1
            stats['datasets'][label]['query_failure'] += 1
            logger.exception(f"Failed to run query",
                             extra={
                                 "source": "service",
                                 'request_id': request_id,
                                 "user": params.get('user', "unknown"),
                                 "dataset": params.get('dataset', "unknown"),
                             })
            query_update_result(request_id, {
                "status": "failure",
                "answer": f"Unable to construct the answer. Could be an internal agent error. See the agent log",
                "result": {
                    "text": str(e)
                }
            })

    async def qna_status(self, request_id: str):
        """
        Get the status of the query request

        Parameters
        ----------
        request_id: str
                    UUID generated for the request

        Returns
        -------
        status: dict
             Dictionary with request id, query, status, message and data
        """

        # First get the params
        try:

            cache = get_cache()

            if request_id not in cache:
                logger.error(f"Query Status: Invalid Request ID",
                             extra={
                                 'request_id': request_id,
                                 "source": "service",
                             })
                return {
                    "request_id": request_id,
                    "status": "failure",
                    'answer': "Data not available. Either the request is invalid or the memory with that request has been cleared on restart"
                }

            value: dict = cache[request_id]
            status = value['status']
            logger.debug(f"Query Status: {status}",
                         extra={
                             "source": "service",
                             'request_id': request_id,
                             "dataset": value.get('dataset', 'unknown'),
                             "user": value.get('user', 'unknown'),
                             'data': json.dumps(value, indent=4, cls=SafeEncoder)
                         })

            return {
                "request_id": request_id,
                "query": value['query'],
                "status": value['status'],
                'message': value.get('message', ""),
                'data': value.get('result', {})
            }
        except Exception as e:
            return {
                "request_id": request_id,
                "query": request_id,
                "status": "failure",
                "message": str(e),
                "data": {}
            }

    async def qna(self, user: str,
                  query: str,
                  dataset: str,
                  background_tasks: BackgroundTasks,
                  mode: str = 'economy',
                  context: str = "",
                  policy: dict = {},
                  extra: dict = {}
                  ):

        """
        Run the query against the LLM microservices

        :param user: User
        :param query:  Query
        :param dataset: Dataset
        :param background_tasks:
        :param mode: Type of mode
        :param context:
        :param policy:
        :param extra:
        :return:
        """
        cache = get_cache()
        request_id = str(uuid4())

        params = {
            "user": user,
            "dataset": dataset,
            "context": context,
            "namespace": self.namespace,
            "query": query,
            "policy": policy,
            "mode": mode
        }

        cache[request_id] = {
            'query': query,
            "status": "pending",
            "user": user,
            "dataset": dataset,
            "params": params
        }

        # Run the background task...
        background_tasks.add_task(self.qna_run, request_id)

        return {
            "request_id": request_id,
            "status": "pending",
            "query": query,
            "data": {}
        }

    async def app_startup(self):
        """
        Start the LLM microservices
        :return:
        """
        logger.info("Starting agent [" + self.agentname + "] with namespace [" + self.namespace + "]...")
        if self.agent is None:
            details = await self.get_llm_agent()
            if details is None:
                query_update_result("request_id", {
                    'status': "failure",
                    "message": f"LLM QueryMapper Agent could not be found/built",
                    "result": {}
                })
                return

            self.agent = details['agent']

        self.startup_extra_steps()



import os
from . import policyapp, singlepageapp

def get_template_dirs():
    """
    Template search paths..
    """
    thisdir = os.path.dirname(os.path.abspath(__file__))
    return [
        os.path.join(thisdir, 'sharedapp', 'templates'),
        os.path.join(thisdir, 'policyapp', 'templates'),
        os.path.join(thisdir, 'singlepageapp', 'templates')
    ]



"""

Transforms have numerous data dependencies, and as time is progressing
the dependencies are increasing and they have to tracked. Further
datasets are used in multiple places including the feature marketplace
and compliance modules.

Abstract dataset specification. This is useful for:

(a) handling dependent datasets
(b) cleaning
(c) checking
(d) validating & sampling
(e) discovery of data

There are two abstractions: Dataset and DatasetRegistry.

"""

from .discover import *
from . import discover
from .doodle import Doodle, TransformDoodle



import re
import json
from datetime import date, datetime
import numpy as np

class SafeEncoder(json.JSONEncoder):

    def default(self, obj):
        try:
            if isinstance(obj, (datetime, date)):
                result = obj.isoformat()
            elif isinstance(obj, (tuple)):
                result = super().default(list(obj))
            elif isinstance(obj, np.integer):
                return super().default(int(obj))
            elif isinstance(obj, float) and np.isnan(obj):
                return "null"
            elif isinstance(obj, np.floating):
                return float(obj)
            elif isinstance(obj, np.ndarray):
                return obj.tolist()
            else:
                result = super().default(obj)
        except:
            result = str(obj)

        return result

def slugify(s):
  s = s.lower().strip()
  s = re.sub(r'[^\w\s-]', '', s)
  s = re.sub(r'[\s_-]+', '_', s)
  s = re.sub(r'^-+|-+$', '', s)
  return s



import traceback

# Disable logging of dependent modules
import logging

logging.getLogger("urllib3").setLevel(logging.CRITICAL)
logging.getLogger("boto").setLevel(logging.CRITICAL)
logging.getLogger("boto3").setLevel(logging.CRITICAL)
logging.getLogger("botocore").setLevel(logging.CRITICAL)
logging.getLogger("nose").setLevel(logging.CRITICAL)
logging.getLogger("fsspec").setLevel(logging.WARNING)
logging.getLogger("s3fs").setLevel(logging.CRITICAL)
logging.getLogger("elasticsearch").setLevel(logging.CRITICAL)
logging.getLogger("elasticsearch.trace").setLevel(logging.CRITICAL)
logging.getLogger("numba.core.byteflow").setLevel(logging.CRITICAL)
logging.getLogger("numba.core.ssa").setLevel(logging.CRITICAL)
logging.getLogger("numba.core.interpreter").setLevel(logging.CRITICAL)

from .node import *
from .sdk import *
from .render import *
from .mixins import *
from .run import *
from .patch import *



import os
import json
from json import JSONEncoder
import pickle
import numpy as np
import pandas as pd
import time
import sqlite3
import tempfile
import hashlib
from enrichsdk import Compute, S3Mixin
from datetime import datetime, timedelta, date
import logging
from sqlalchemy import create_engine, text as satext

import great_expectations as ge
from great_expectations.data_context.types.base import DataContextConfig, DatasourceConfig
from great_expectations.data_context import BaseDataContext
from great_expectations.data_context.types.base import FilesystemStoreBackendDefaults
from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.render.renderer import *
from great_expectations.render.view import DefaultJinjaPageView

from enrichsdk.contrib.lib.transforms import note
from enrichsdk.contrib.lib.assets import profilespec
from enrichsdk.datasets import TransformDoodle
from enrichsdk.utils import SafeEncoder

logger = logging.getLogger("app")

class CustomGEJSONEncoder(SafeEncoder):
    def default(self, o):
        try:
            return super().default(o)
        except:
            if hasattr(o, "__dict__"):
                return o.__dict__
            else:
                return str(o)

class DataQualityBase(Compute):
    """
    Run data quality checks against a data source based on a spec

    Features of transform baseclass include:
        * Flexible configuration
        * Highlevel specification of observability:
            * specified data source
            * custom defined data quality checks (same DSL as Great Expectation python package)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "DataQualityBase"
        self.description = "Data quality checks for a data source given a spec"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

    @classmethod
    def instantiable(cls):
        return False

    def get_handlers(self, spec):
        """
        Define various callbacks that take a dataframe, spec
        and compute.
        """
        return {}

    ###########################################
    # Process the expectations spec
    ###########################################
    def process_spec(self, spec, data):

        if data is None:
            msg = "No dataset loaded" + "\n"
            logger.exception(
                f"Spec: {spec['name']} -- skipping",
                extra={"transform": self.name}
            )
            return None

        # setup a GE object
        df_ge = ge.from_pandas(data)

        name = spec['name']
        config = spec['config']
        expectations = config['expectations']

        # what is the expected result format?
        result_format = config.get('result_format', 'none')
        if result_format not in ["basic", "complete", "boolean_only", "summary"]:
            result_format = "basic"
        result_format = result_format.upper()

        tested_at = datetime.now()

        run_name = f"{name}-{tested_at}"
        run_name = hashlib.md5(run_name.encode('utf-8')).hexdigest()

        # check if checks object is a list
        if not isinstance(expectations, list):
            logger.exception(
                f"Invalid config param -- expectations", extra={"transform": self.name, "data": json.dumps(config, indent=4)}
            )
            return None

        # for each expectation to test
        for expectation in expectations:
            do_test = True

            for f in ['column', 'expectation', 'parameters']:
                if f not in expectation:
                    logger.exception(
                        f"Invalid config param -- expectations", extra={"transform": self.name, "data": json.dumps(config, indent=4)}
                    )
                    do_test = False
                    break
            if do_test is False:
                continue

            # set some params explicitly for the call to GE
            expectation['parameters']['column'] = expectation['column']
            expectation['parameters']['result_format'] = result_format

            # now we can run the expectation test
            if expectation.get('type', 'builtin') == 'builtin':
                method = expectation['expectation']
                params = expectation['parameters']

                if not hasattr(df_ge, method):
                    logger.exception(
                        f"Invalid expectation method", extra={"transform": self.name, "data": json.dumps(expectation, indent=4)}
                    )
                    continue

                # we now have the callback
                callback = getattr(df_ge, method)
                try:
                    r = callback(**params)
                except:
                    logger.exception(
                        f"Error in expectation method", extra={"transform": self.name, "data": json.dumps(expectation, indent=4)}
                    )
                    continue

            elif expectation.get('type', 'builtin') == 'custom':
                pass
            else:
                logger.exception(
                    f"Unknown expectation type", extra={"transform": self.name, "data": json.dumps(expectation, indent=4)}
                )
                continue


        # generate the validataions report
        result_ge = df_ge.validate(run_id=run_name)

        msg = f"Total expectations tested: {len(result_ge['results'])}" + "\n"
        msg += f"{result_ge}"
        logger.debug(
            f"Processed expectations: {spec['name']}",
            extra={"transform": self.name, "data": msg}
        )

        return result_ge


    ###########################################
    # Helper Functions
    ###########################################

    def update_frame(self, source, description, df, lineage=None):
        if isinstance(source, str):
            name = source
        else:
            name = source["name"]

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    ###########################################
    # I/O Functions
    ###########################################

    def read_s3_data(self, filename, params={}):
        # assume we have a resolved s3fs object
        s3 = self.args['s3']
        if s3.exists(filename):
            df = pd.read_csv(s3.open(filename), **params)
            return df
        return None

    def get_dataset_s3(self, spec, paths):
        '''
        Gets all files from paths and puts them together
        into a single dataframe. If self.args['cache']==True,
        then this consolidated dataframe is cached / read from cache
        as applicable.
        '''
        msg = ""

        run_date    = self.args['run_date']
        config      = spec['config']
        dataset     = config['dataset']

        cache = self.args.get("cache", False)
        cachename = f"{dataset}-{run_date}"
        cachefile = f"cache/{self.name}-rawdata-cache-" + cachename + ".csv"

        # read from cache if available
        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                msg = f"Location: {cachefile}" + "\n"
                df = pd.read_csv(cachefile)
                logger.debug(f"Read cached {dataset}", extra={"transform": self.name})
                return df

        # read from S3
        dfs = []
        for path in paths:
            _df = self.read_s3_data(path)
            if _df is None:
                msg += f"Path not found, skipping: {path}" + "\n"
                continue
            msg += f"Read from path: {path}" + "\n"
            dfs.append(_df)
        df = pd.concat(dfs)

        logger.debug(f"Read fresh {dataset}", extra={"transform": self.name})

        # Cache it for future use
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if (len(paths) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": [paths],
                    },
                ],
            }

        if not self.state.has_frame(spec['name']):
            self.update_frame(spec, f"Dataset: {dataset}", df, lineage)

        return df


    def get_dataset_db(self, spec):
        name = spec['name']
        config = spec['config']
        source = config['source']

        testmode = self.args.get('testmode', False)
        env      = 'test' if testmode else 'prod'

        # Get the db engine
        engine_name = source['params'][env]['engine']
        engine = self.engines[engine_name]

        # Get the query
        query = source['params'][env]['query']
        if query.endswith('.sql'):
            # we have a file containing the query
            sqlfile = os.path.join(self.scriptdir, query)
            query = open(sqlfile).read()

        df = self.read_db_source(engine, query)

        # Now the input load...
        lineage = {
            "type": "lineage",
            "transform": self.name,
            "dependencies": [
                {
                    "type": "database",
                    "nature": "input",
                    "objects": [source['dataset']],
                },
            ],
        }

        self.update_frame(spec, f"Dataset: {name}", df, lineage)

        return df

    def read_db_source(self, engine, query):
        # Run the query
        df = pd.read_sql_query(satext(query), con=engine)

        return df

    def load_dataset(self, spec):
        name        = spec['name']
        config      = spec['config']

        source_name = config['dataset']
        source_id   = config['source_id']
        source_version = config.get('source_version', 'v1')
        dataset     = config['dataset']

        datacred = self.args['datacred']
        doodle = TransformDoodle(self, self.state, datacred)

        source, paths = doodle.get_source_paths(start=datetime.today() + timedelta(days=-7),
                                                end=datetime.today(),
                                                name=source_name,
                                                version=source_version,
                                                source_id=source_id)

        # Insert a read action..
        try:
            result = doodle.update_source(source['id'], {})
            logger.debug(f"Updated doodle: {source_name}",
                         extra={
                             'transform': self.name,
                             'data': json.dumps(result, indent=4, cls=SafeEncoder)
                         })
        except:
            logger.exception(f"Unable to update doodle {source_name}",
                             extra={
                                 'transform': self.name
                             })

        df = self.get_dataset_s3(spec, paths)

        msg = note(df, f"Input Dataset: {dataset}")
        logger.debug(
            f"Loaded dataset for spec: {name}", extra={"transform": self.name, "data": msg}
        )

        return df

    def s3_store_result(self, spec, result):
        specid      = spec['id']
        appname     = spec.get('app','common')
        name        = spec['name']
        namespace   = spec['namespace']
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = time.time()


        # where are we storing it?
        targetdir = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")

        resultfile = os.path.join(targetdir, f"health.json")
        metadatafile = os.path.join(targetdir, f"metadata.json")

        metadata = self.get_default_metadata(self.state)
        metadata['spec'] = spec

        # write to s3
        s = json.dumps(result, indent=4, cls=CustomGEJSONEncoder)
        with s3.open(resultfile, 'w') as fd:
            fd.write(json.loads(s))
        with s3.open(metadatafile, 'w') as fd:
            json.dump(metadata, fd, indent=4, cls=SafeEncoder)

        msg = f"s3 location: {resultfile}" + "\n"
        msg += f"metadata location: {metadatafile}" + "\n"

        logger.debug(f"Wrote data quality results to S3",
                        extra={"transform": self.name,
                                "data": msg})

    def db_store_result(self, spec, result):
        name    = spec['name']
        config  = spec['config']
        store   = config['store']

        testmode = self.args.get('testmode', False)
        env      = 'test' if testmode else 'prod'

        # Get the db engine
        engine_name = store['params'][env]['engine']
        engine = self.engines[engine_name]

        # write results to db
        table_name = store['params'][env]['table']
        ret = results.to_sql(table_name,
                            engine,
                            if_exists='append',
                            index=False)
        # write data to db
        data_table_name = name.replace('-','_')
        ret = data.to_sql(data_table_name,
                            engine,
                            if_exists='append',
                            index=False)

        msg = f"DB engine: {engine}" + "\n"

        logger.debug(f"Wrote monitor results to DB",
                        extra={"transform": self.name,
                                "data": msg})

    def store_result(self, spec, result):
        name    = spec['name']
        config  = spec['config']
        store   = config.get('store', {"sink": "s3"})

        for f in ["sink"]:
            if f not in store:
                logger.exception(
                    f"Store has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                return

        sink = store['sink']
        if sink == "s3":
            # store in s3
            self.s3_store_result(spec, result)
        elif sink == "db":
            # store in db
            self.db_store_result(spec, result)
        else:
            logger.exception(f"Unknown store for dataset: {name}",
                         extra={
                             'transform': self.name
                         })

    ###########################################
    # Expectation Testing Functions
    ###########################################

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "Start execution", extra=self.config.get_extra({"transform": self.name})
        )

        # Will be used in other places..
        self.state = state

        # Get the profile spec
        is_valid, profile, msg = profilespec.get_profile(self, "policyapp.dataqualityv2")
        if is_valid:
            logger.debug(
                f"Loaded profilespec",
                extra={"transform": self.name, "data": msg}
            )
        else:
            logger.error(
                f"Could not load profilespec",
                extra={"transform": self.name, "data": msg}
            )
            raise Exception("could not load profilespec")

        specs = profile.get("specs", None)
        if specs is None:
            raise Exception("Could not find 'specs' in profile")

        # Now go through each spec and process it
        for spec in specs:

            ## first, some checks on the spec
            do_process_spec = True
            name = spec.get('name', 'NO_SPEC_NAME')

            enabled = spec.get("active", True)
            if not enabled:
                logger.debug(
                    f"Spec not enabled, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                do_process_spec = False
                continue

            for f in ["name", "config"]:
                if f not in spec:
                    logger.exception(
                        f"Spec has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue

            config = spec['config']

            for f in ["source_id", "expectations"]:
                if f not in config:
                    logger.exception(
                        f"Spec config has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue

            ## we can now proceed with processing the spec
            # frist, load the source data
            data = self.load_dataset(spec)

            # then, process it
            result = self.process_spec(spec, data)
            if result is None:
                continue

            ## store the expectation validation result
            self.store_result(spec, result)


        # Done
        logger.debug(
            "Complete execution", extra=self.config.get_extra({"transform": self.name})
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass



import os
import subprocess
import uuid
import hashlib



def get_checksum(path):

    if not os.path.isfile(path) or not os.path.exists(path):
        return None

    sha256_hash = hashlib.sha256()
    with open(path,"rb") as f:
        for byte_block in iter(lambda: f.read(4096),b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def get_file_size(path):

    if not os.path.isfile(path) or not os.path.exists(path):
        return 0

    statinfo = os.stat(path)
    return statinfo.st_size


def get_consistent_uuid(text):
    # https://stackoverflow.com/questions/41186818/how-to-generate-a-random-uuid-which-is-reproducible-with-a-seed-in-python

    for n in range(0, 10):
        m = hashlib.md5()
        m.update((text + " " * n).encode("utf-8"))
        new_uuid = uuid.UUID(m.hexdigest())
        if str(new_uuid)[0] == "0":
            continue
        else:
            break

    assert str(new_uuid)[0] != "0"
    return new_uuid



import os
import io
import re
import json
import numpy as np
import pandas as pd
import time
from enrichsdk import Compute, S3Mixin
from datetime import datetime, timedelta, date
from scipy.spatial import distance
import seaborn as sns
import matplotlib.pyplot as plt
import logging

from pandas.api.types import is_numeric_dtype

from enrichsdk.contrib.lib.transforms import note
from enrichsdk.contrib.lib.assets import profilespec
from enrichsdk.datasets import TransformDoodle
from enrichsdk.datasets import DynamicCustomDataset
from enrichsdk.utils import SafeEncoder

logger = logging.getLogger("app")


class AnomaliesBase(Compute):
    """
    Compute anomalies given a dataframe with columns

    Features of transform baseclass include:

        * Flexible configuration
        * Highlevel specification of columns combinations and detection strategy

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "AnomaliesBase"
        self.description = "Compute anomalies in column(s) of a dataframe"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

    @classmethod
    def instantiable(cls):
        return False

    def get_handlers(self, spec):
        """
        Define various callbacks that take a dataframe, spec
        and compute.
        """
        return {}

    def get_profile(self):
        """
        Read the profile json from API
        """

        if (not hasattr(self, "args")):
            raise Exception(
                "'args' transform attribute should be defined to use default get_profile method"
            )
        for p in ['apicred']:
            if self.args.get(p) == None:
                raise Exception(
                    f"'{p}' attribute in args should be defined to use default get_profile method"
                    )

        # call the API to get the anomaly specs
        anomalyspecs, is_valid, msg = load_profile_api(self.args)
        logger.debug(
            f"Loading profile from API",
            extra={"transform": self.name, "data": msg},
        )
        if is_valid == False:
            raise Exception(f"Error loading profile")

        specs = anomalyspecs["specs"]
        logger.debug(
            f"Found {len(specs)} specs",
            extra={"transform": self.name, "data": json.dumps(anomalyspecs, indent=4)},
        )

        return anomalyspecs

    def get_specs(self, profile):
        if (not isinstance(profile, dict)) or ("specs" not in profile):
            raise Exception("Specs not defined in profile")
        return profile["specs"]

    def preprocess_spec(self, spec):
        '''
        to be overloaded in the derived class
        '''
        return spec

    def process_spec(self, spec, data):
        msg = ""

        if data is None:
            msg = "No dataset loaded" + "\n"
            logger.exception(
                f"Spec: {spec['name']} -- skipping",
                extra={"transform": self.name}
            )
            return None

        name = spec["name"]
        config = spec["config"]
        handlers = self.get_handlers(spec)

        # use the specified strategy to detect anomalies
        try:
            strategy = config.get("strategy")
            if strategy == None:
                # We are doing the default anomaly detection which is deviation from mean
                data = self.process_spec_default(data, spec)
            elif strategy == "custom":
                # Custom callback
                data  = self.process_spec_custom(data, spec)
            else:
                # for now, we fallback to the default strategy
                data  = self.process_spec_default(data, spec)

            # The spec processor can return multiple dataframes
            if isinstance(data, pd.DataFrame):
                data = {name: data, "stats": {}}
        except:
            logger.exception(f"Failed to process {name}",
                             extra={
                                 'transform': self.name,
                                 'data': msg
                             })

            return None

        logger.debug(
            f"Processed anomalies",
            extra={"transform": self.name, "data": msg}
        )

        return data

    def process_spec_custom(self, datasets, profile, spec):
        msg = ""

        name = spec["name"]
        handlers = self.get_handlers(profile)

        # Custom...
        generate = spec["generate"]
        callback = None
        if (generate in handlers) and (not callable(handlers[generate])):
            callback = handlers[generate]
        elif hasattr(self, generate):
            callback = getattr(self, generate)

        if callback is None:
            raise Exception(f"[{name}] Invalid callback: {generate}")

        # Get hold of the data first...
        sources = self.get_spec_sources(spec, datasets)

        return callback(sources, spec)

    def process_spec_default(self, data, spec):
        """
        Handle one specification at a time..
        """

        partialsamplerate = 0.05
        samplerate_lut = {
            "all": 1.0,
            "partial": partialsamplerate,
            "none": 0.0
        }
        tolerances = {
            "low": 1,
            "medium": 2,
            "high": 3,
        }

        def anomaly_note(row, threshold):
            distance = row[f"__anomaly_distance__"]
            if distance > threshold:
                return f"{(round(distance/threshold,2))}x outside expected sample deviation"
            return f"within expected sample deviation"


        msg = ""
        msg += f"Using default centroid distance anomaly detector" + "\n"

        config = spec["config"]
        msg += f"Config: {json.dumps(config, indent=4)}" + "\n"

        # Get hold of the data first...
        name = spec["name"]
        orig_df = data
        total_samples = len(orig_df)

        metrics     = config.get("metrics", orig_df.columns)
        groups      = config.get('groups', [])
        outputs     = config.get("outputs", orig_df.columns)
        dimensions  = config.get("dimensions", orig_df.columns)
        columns     = list(set(metrics + outputs + dimensions))

        msg += f"Combined set of columns: {columns}" + "\n"
        msg += f"{note(orig_df, 'Original DF')}" + "\n"

        #########
        # default anomaly detection
        #########
        # get tolerance thresold
        tolerance = config.get("threshold", config.get("thresold", "medium"))
        scalefactor = tolerances.get(tolerance, 2)

        # get the sample strategy for the normal data
        normal_samples = config.get("normal_samples", "partial")
        samplerate = samplerate_lut[normal_samples]

        msg += f"(tolerance, scalefactor): ({tolerance}, {scalefactor})" + "\n"

        logger.debug(f"Setting up for spec: {spec['name']}",
                         extra={
                             'transform': self.name,
                             'data': msg
                         })

        anomaly_stats = {}
        plotdata = {}
        dfs = []

        #########
        # we first do the leaf level, per metric to check for anomalies
        #########
        msg = f"Processing metrics: {metrics}" + "\n\n"

        for metric in metrics:

            # make a copy of the df, we'll keep adding anomlay metrics to it
            df = orig_df[columns].copy()

            if not is_numeric_dtype(df[metric]):
                msg += f"{metric} Metric not numeric. Skipping\n"
                continue

            # compute the anomalies for this metric
            points      = df[metric].to_numpy()     # all data as an MxN matrix
            centroid    = df[metric].mean()          # the computed centroid of the dataset
            distances   = abs(points - centroid)    # distances of each point to centroid
            stddev      = np.nanstd(points)      # std dev of distances
            threshold   = stddev * scalefactor
            anomalies   = np.where(distances.flatten()>threshold, 'anomaly', 'normal')    # flag where anomalies occur

            # add columns indicating anomaly label
            id = f"metric-{metric}"
            df['id'] = id
            df['level'] = 'metric'
            df['name'] = metric
            df['__is_anomaly__'] = pd.Series(anomalies)

            # add columns indicating reason for anomaly
            df[f"__anomaly_distance__"] = pd.Series(distances.flatten())
            df[f"__anomaly_note__"] = df.apply(lambda x: anomaly_note(x, threshold), axis=1)

            df_a = df[df['__is_anomaly__']=='anomaly']
            n_anomalies = len(df_a)
            perc_anomalies = round(n_anomalies/total_samples*100, 2)

            df_n = df[df['__is_anomaly__']=='normal'].sample(frac=samplerate)
            df_n = df_n[0:min(3*n_anomalies,len(df_n))] # min 3x n_anomalies or configured sample of normal samples
            n_nsamples = len(df_n)

            # for this metric, we now have all the detected anomalies and the sampled normal data
            sampled_df = pd.concat([df_a, df_n])

            msg += f"--------------------------" + "\n"
            msg += f"Metric: {metric}" + "\n"
            msg += f"Computed stddev: {stddev}" + "\n"
            msg += f"Threshold: {threshold}" + "\n"
            msg += f"Anomalies: {n_anomalies}/{total_samples}={perc_anomalies}%" + "\n"
            msg += f"--------------------------" + "\n\n"

            anomaly_stats[id] = {
                "level": 'metric',
                "name": metric,
                "dimensions": dimensions,
                "n_anomalies": n_anomalies,
                "perc_anomalies": perc_anomalies,
                "n_normalsamples": n_nsamples,
                "n_plotsamples": len(df),
            }
            plotdata[id] = df

            dfs.append(sampled_df)

        logger.debug(f"Processed metrics level: {spec['name']}",
                         extra={
                             'transform': self.name,
                             'data': msg
                         })


        # #########
        # # then we do the group level, hierarchial
        # #########
        msg = f"Processing groups: {groups}" + "\n\n"

        for group in groups:
            group_name      = group.get('group')
            g_dimensions    = group.get('dimensions', dimensions)
            g_metrics       = group.get('metrics')

            # we don't have what we need, skip
            if group_name == None or metrics == None:
                continue

            if not all([is_numeric_dtype(df[metric]) for metric in g_metrics]):
                msg += f"{group_name} One or more metrics are not numeric\n"
                continue

            # make a copy of the df, we'll keep adding anomlay metrics to it
            df = orig_df[columns].copy()

            points      = df[g_metrics].to_numpy()    # all data as an MxN matrix
            centroid    = df[g_metrics].mean().values # the computed centroid of the dataset
            distances   = distance.cdist(points, np.array([centroid]), 'euclidean') # distances of each point to centroid
            distances   = np.reshape(distances, len(distances))
            stddev      = np.nanstd(points)         # std dev of distances
            threshold   = stddev * scalefactor
            anomalies   = np.where(distances.flatten()>threshold, 'anomaly', 'normal')    # flag where anomalies occur

            # add columns indicating anomaly label
            id = f"group-{group_name}"
            df['id'] = id
            df['level'] = 'group'
            df['name'] = group_name
            df['__is_anomaly__'] = pd.Series(anomalies)

            # add columns indicating reason for anomaly
            df[f"__anomaly_distance__"] = pd.Series(distances.flatten())
            df[f"__anomaly_note__"] = df.apply(lambda x: anomaly_note(x, threshold), axis=1)

            df_a = df[df['__is_anomaly__']=='anomaly']
            n_anomalies = len(df_a)
            perc_anomalies = round(n_anomalies/total_samples*100, 2)

            df_n = df[df['__is_anomaly__']=='normal'].sample(frac=samplerate)
            df_n = df_n[0:min(3*n_anomalies,len(df_n))] # min 3x n_anomalies or configured sample of normal samples
            n_nsamples = len(df_n)

            # for this metric, we now have all the detected anomalies and the sampled normal data
            sampled_df = pd.concat([df_a, df_n])

            msg += f"--------------------------" + "\n"
            msg += f"Group: {group_name}" + "\n"
            msg += f"Computed stddev: {stddev}" + "\n"
            msg += f"Threshold: {threshold}" + "\n"
            msg += f"Anomalies: {n_anomalies}/{total_samples}={perc_anomalies}%" + "\n"
            msg += f"--------------------------" + "\n"

            anomaly_stats[id] = {
                "level": 'group',
                "name": group_name,
                "metrics": g_metrics,
                "dimensions": g_dimensions,
                "threshold": threshold,
                "n_anomalies": n_anomalies,
                "perc_anomalies": perc_anomalies,
                "n_normalsamples": n_nsamples,
                "n_plotsamples": len(df),
            }
            plotdata[id] = df

            dfs.append(sampled_df)

        logger.debug(f"Processed groups level: {spec['name']}",
                         extra={
                             'transform': self.name,
                             'data': msg
                         })

        if len(dfs) == 0:
            logger.debug(f"{name}: No outputs computed",
                         extra={
                             'transform': self.name,
                             'data': msg
                         })
            return None

        #########
        # construct the DF for output
        #########
        # concat for all metrics+groups
        df = pd.concat(dfs)
        # reorder columns
        first_cols = ['id', 'level', 'name']
        cols = first_cols + [c for c in df.columns if c not in first_cols]
        df = df[cols]

        msg = f"Final columns: {df.columns}" + "\n"

        window, start_date, end_date = self.get_window_dates(config, self.args)

        # compute stats of interest
        stats = {
            "timestamp": f"{datetime.now().isoformat()}",
            "policy": config,
            "data_start_date": f"{start_date}",
            "data_end_date": f"{end_date}",
            "strategy": "centroid",
            "tolerance": tolerance,
            "scalefactor": scalefactor,
            "normalsamples": normal_samples,
            "samplerate": samplerate,
            "n_rows": total_samples,
            "anomaly_stats": anomaly_stats,
        }

        msg += f"Stats: {json.dumps(stats, indent=4)}" + "\n"

        msg += f"{note(df, 'Anomaly DF')}" + "\n"

        logger.debug(f"Completed spec: {spec['name']}",
                         extra={
                             'transform': self.name,
                             'data': msg
                         })

        return {name: df, "stats": stats, "plotdata": plotdata}

    def summarize_dimension(self, df, d):
        col = '__is_anomaly__'
        max_cats = 10

        # first get the category labels for anomalies, we need to retain these
        anomaly_cats = list(set(df[df[col]=='anomaly'][d].values))

        # remove these rows, so they don't interfere in the next steps
        _df = df[df[col]!='anomaly'].copy()

        # determine how many topk categories to keep
        # if we already have too many categories from the anomaly rows
        # this will be 0
        topk = max(max_cats - len(anomaly_cats), 0)

        # then, get the top-K represented categories
        topk_cats = _df[[d]+[col]] \
                    .groupby(by=[d]) \
                    .agg('count') \
                    .sort_values(by=col, ascending=False) \
                    .head(topk) \
                    .reset_index()[d] \
                    .tolist()

        # what categoeies are we retaining
        retain_cats = anomaly_cats + topk_cats

        # label every other category as OTHER
        df[d] = df.apply(lambda x: 'OTHER' if x[d] not in retain_cats else x[d],
                        axis=1)

        return df


    def read_db_source(self, source):

        # Get the SQLAlchemy URI
        uri = self.get_db_uri(source)

        # Get the query
        query = self.get_db_query(source)

        # Create the engine
        engine = create_engine(uri)

        # Run the query
        df = pd.read_sql_query(satext(query), con=engine)

        # Now the input load...
        lineage = {
            "type": "lineage",
            "transform": self.name,
            "dependencies": [
                {
                    "type": "database",
                    "nature": "input",
                    "objects": [self.get_printable_db_uri(uri)],
                },
            ],
        }

        self.update_frame(source, f"Dataset: {dataset}", df, lineage)

        return df

    def read_s3_data(self, filename, params={}):
        # assume we have a resolved s3fs object
        s3 = self.args['s3']
        if s3.exists(filename):
            df = pd.read_csv(s3.open(filename), **params)
            return df
        return None

    def get_spec_sources(self, spec, datasets):

        name = spec["name"]

        if ("sources" not in spec) and ("source" not in spec):
            raise Exception(f"[{name}] Invalid specification. Missing dataset")

        sources = spec.get("sources", spec.get("source"))
        if isinstance(sources, str):
            sources = [sources]

        policy = spec.get("missing", "fail")
        for s in sources:
            if s not in datasets:
                if policy == "fail":
                    raise Exception(f"[{name}] Missing source: {s}")

        return {s: datasets[s] for s in sources if s in datasets}

    def update_frame(self, source, description, df, lineage=None):

        if isinstance(source, str):
            name = source
            description = ""
        else:
            name = source["name"]
            description = source.get("description", "")

        if self.state.has_frame(name):
            return

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    def load_dataset(self, spec):
        name        = spec['name']
        config      = spec['config']

        source_name = config['dataset']
        source_id   = config['source_id']
        source_version = config.get('source_version', 'v1')
        dataset     = config['dataset']

        datacred = self.args['datacred']
        doodle = TransformDoodle(self, self.state, datacred)

        #=> the source may have a different id than source_id
        # because we gave both the name and the id
        source, paths = doodle.get_source_paths(start=datetime.today(), # + timedelta(days=-7),
                                                end=datetime.today(),
                                                name=source_name,
                                                version=source_version,
                                                source_id=source_id)

        # Insert a read action..
        try:
            result = doodle.update_source(source['id'], {})
            logger.debug(f"Updated doodle: {source_name}",
                         extra={
                             'transform': self.name,
                             'data': json.dumps(result, indent=4, cls=SafeEncoder)
                         })
        except:
            logger.exception(f"Unable to update doodle {source_name}",
                             extra={
                                 'transform': self.name
                             })


        df = self.get_dataset_s3(spec, paths)

        msg = note(df, f"Input Dataset: {dataset}")
        logger.debug(
            f"Loaded dataset for spec: {name}", extra={"transform": self.name, "data": msg}
        )

        return df

    def get_window_dates(self, config, args):
        # get the window size
        window = config.get("window")
        if window == None or window == "":
            window = "1"
        try:
            window = int(window)
        except:
            raise Exception(
                "window param in config needs to be a string integer"
            )

        # determine start and end dates for dataset
        end_date = args["run_date"]
        start_date = end_date + timedelta(days=-window+1)

        return window, start_date, end_date


    def get_dataset_s3(self, spec, paths):
        '''
        Gets all files from paths and puts them together
        into a single dataframe. If self.args['cache']==True,
        then this consolidated dataframe is cached / read from cache
        as applicable.
        '''
        msg = ""

        run_date    = self.args['run_date']
        config      = spec['config']
        dataset     = config['dataset']

        cache = self.args.get("cache", False)
        cachename = f"{dataset}-{run_date}"
        cachefile = f"cache/{self.name}-rawdata-cache-" + cachename + ".csv"

        # read from cache if available
        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                msg = f"Location: {cachefile}" + "\n"
                df = pd.read_csv(cachefile)
                logger.debug(f"Read cached {dataset}", extra={"transform": self.name})
                return df

        # read from S3
        dfs = []
        for path in paths:
            _df = self.read_s3_data(path)
            if _df is None:
                msg += f"Path not found, skipping: {path}" + "\n"
                continue
            msg += f"Read from path: {path}" + "\n"
            dfs.append(_df)
        df = pd.concat(dfs)

        logger.debug(f"Read fresh {dataset}", extra={"transform": self.name})

        # Cache it for future use
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if (len(paths) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": [paths],
                    },
                ],
            }

        if not self.state.has_frame(spec['name']):
            self.update_frame(spec, f"Dataset: {dataset}", df, lineage)

        return df

    def save_image(self, fig, plotfile, testplotfile):
        s3 = self.args['s3']

        # save locally when testing
        if self.args.get('testmode', False):
            outputpath = self.args['test_outpath']
            try:
                os.makedirs(outputpath)
            except:
                pass
            testplotfile = os.path.join(outputpath, testplotfile)
            fig.savefig(testplotfile)

        # write to s3
        img_data = io.BytesIO()
        fig.savefig(img_data, format='png')
        img_data.seek(0)
        with s3.open(plotfile, 'wb') as fd:
            fd.write(img_data.getbuffer())


    def s3_store_result(self, spec, data):
        name        = spec['name']
        appname     = spec.get('app','common')
        namespace   = spec.get('namespace', 'default')
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = time.time()

        config      = spec["config"]

        # get the dataframe and
        # add additional columns
        df = data[name]
        df["__run_date__"] = run_date

        # get the stats object
        stats = data["stats"]
        plotdata = data["plotdata"]

        # where are we storing it?
        targetdir = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")

        anomaliesfile   = os.path.join(targetdir, f"outliers.csv")
        statsfile       = os.path.join(targetdir, f"stats.json")
        metadatafile    = os.path.join(targetdir, f"metadata.json")

        metadata = self.get_default_metadata(self.state)
        metadata['spec'] = spec

        ## plot the results
        msg = note(df, f"{name} - Anomalies Tagged") + "\n"
        msg += f"Anomalies: {anomaliesfile}" + "\n"
        msg += f"Stats: {statsfile}" + "\n"

        # for each level-name combination (indexed by id)
        for id in set(df['id'].values):

            data_df = plotdata[id]
            data_df = data_df.reset_index(drop=True)

            # Plot the data
            if stats['anomaly_stats'][id]['level'] == 'metric':
                # for the metrics level
                x = stats['anomaly_stats'][id]['name']
                dimensions = stats['anomaly_stats'][id]['dimensions']
                plot = sns.displot(data_df,
                                    x=x,
                                    hue="__is_anomaly__",
                                    stat="probability",
                                    bins=100)
                plot.fig.subplots_adjust(top=.95)
                plot.set(title=f'Metric Distribution ({id})')
                fig = plot.fig
                testplotfile = f'{name}-{id}.png'
                plotfile = os.path.join(targetdir, f"plot-{id}.png")
                msg += f"plot: {plotfile}" + "\n"
                self.save_image(fig, plotfile, testplotfile)

                # for each dimension, we show a categorical plot
                for d in dimensions:
                    if data_df[d].dtype not in ['O', 'string', 'str']:
                        # pass if numeric type
                        continue
                    m_data_df = data_df[[d]+[x]+['__is_anomaly__']] # keep only the columns we need
                    m_data_df = self.summarize_dimension(data_df, d)
                    plot = sns.catplot(
                                data=m_data_df, kind="strip",
                                x=x, y=d, hue="__is_anomaly__")
                    plot.fig.subplots_adjust(top=.95)
                    plot.set(title=f'Distribution of {d} ({id})')
                    fig = plot.fig
                    testplotfile = f'{name}-{id}-dimension-{d}.png'
                    plotfile = os.path.join(targetdir, f"plot-{id}-dimension-{d}.png")
                    msg += f"plot: {plotfile}" + "\n"
                    self.save_image(fig, plotfile, testplotfile)

            else:
                # for the groups level
                metrics = stats['anomaly_stats'][id]['metrics']
                dimensions = stats['anomaly_stats'][id]['dimensions']
                data_df = data_df[dimensions+metrics+['__is_anomaly__']]  # only retain the columns we need
                plot = sns.pairplot(data_df,
                                    hue='__is_anomaly__')
                plot.fig.subplots_adjust(top=.95)
                plot.fig.suptitle(f'Pairplot Structure ({id})')
                fig = plot.fig

                testplotfile = f'{name}-{id}.png'
                plotfile = os.path.join(targetdir, f"plot-{id}.png")
                msg += f"plot: {plotfile}" + "\n"
                self.save_image(fig, plotfile, testplotfile)


        # write data to s3
        with s3.open(anomaliesfile, 'w') as fd:
            df.to_csv(fd, index=False)
        with s3.open(statsfile, 'w') as fd:
            json.dump(stats, fd, indent=4, cls=SafeEncoder)
        with s3.open(metadatafile, 'w') as fd:
            json.dump(metadata, fd, indent=4, cls=SafeEncoder)

        logger.debug("Wrote anomalies".format(name),
                        extra={"transform": self.name,
                                "data": msg})

    def store_result(self, spec, data):
        name    = spec['name']
        config  = spec['config']
        store   = config.get('store', {"sink": "s3"})

        for f in ["sink"]:
            if f not in store:
                logger.exception(
                    f"Store has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                return

        sink = store['sink']
        if sink == "s3":
            # store in s3
            self.s3_store_result(spec, data)
        elif sink == "db":
            # store in db
            self.db_store_result(spec, data)
        else:
            logger.exception(f"Unknown store for dataset: {name}",
                         extra={
                             'transform': self.name
                         })

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "Start execution", extra=self.config.get_extra({"transform": self.name})
        )

        # Will be used in other places..
        self.state = state

        # Get the anomaly profile
        is_valid, profile, msg = profilespec.get_profile(self, "policyapp.outliersv2")
        if is_valid:
            logger.debug(
                f"Loaded profilespec",
                extra={"transform": self.name, "data": msg}
            )
        else:
            logger.error(
                f"Could not load profilespec",
                extra={"transform": self.name, "data": msg}
            )
            raise Exception("could not load profilespec")

        specs = profile.get("specs", None)
        if specs is None:
            raise Exception("Could not find 'specs' in profile")

        # Now go through each spec and generate anomaly reports
        for spec in specs:

            ## first, some checks on the spec
            do_process_spec = True
            name = spec.get('name', 'NO_SPEC_NAME')

            enabled = spec.get("active", True)
            if not enabled:
                logger.debug(
                    f"Spec not enabled, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                do_process_spec = False
                continue

            for f in ["name", "config"]:
                if f not in spec:
                    logger.exception(
                        f"Spec has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue

            config = spec['config']

            for f in ["source_id"]:
                if f not in config:
                    logger.exception(
                        f"Spec config has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break

            if not do_process_spec:
                continue

            ## pre-process the spec
            try:
                spec = self.preprocess_spec(spec)
                logger.debug(f"Preproccessed spec: {spec['name']}",
                             extra={
                                 'transform': self.name,
                                 'data': json.dumps(spec, indent=4)
                             })

                ## we can now proceed with processing the spec
                # frist, load the source data
                data = self.load_dataset(spec)

                ## process the spec to detect outliers
                data = self.process_spec(spec, data)

                if ((not isinstance(data, dict)) or
                    (len(data) == 0)):
                    continue

                # write the detected outliers
                self.store_result(spec, data)
            except:
                logger.exception(f"Failed to process {name}",
                                 extra={
                                     'transform': self.name
                                 })

        # Done
        logger.debug(
            "Complete execution", extra=self.config.get_extra({"transform": self.name})
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass



import io
import os
import json
import math
import pickle
import numpy as np
import pandas as pd
import time
import sqlite3
import tempfile
import hashlib
from enrichsdk import Compute, S3Mixin
from datetime import datetime, timedelta, date
import logging
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.patches import Rectangle

from enrichsdk.contrib.lib.transforms import note
from enrichsdk.contrib.lib.assets import profilespec
from enrichsdk.contrib.lib.assets import BaseChangePointDetectorModel
from enrichsdk.datasets import TransformDoodle
from enrichsdk.utils import SafeEncoder

from dateutil import parser as dateparser

logger = logging.getLogger("app")

class ChangePointDetectorBase(Compute):
    """
    Take a timeseries signal and identify changepoints in the signal

    Features of transform baseclass include:
        * Flexible configuration
        * Highlevel specification of change point detection:
            * specified data source or custom method to generate one
            * generic change point detection method or custom defined ones
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "ChangePointDetectorBase"
        self.description = "Change point(s) detection for a timeseries signal given a spec"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

        self.cautions = {
            "low": {"color": "green", "desc": "MINIMAL", "rec": "In a BUSINESS-AS-USUAL regime now."},
            "medium": {"color": "gold", "desc": "LOW to MODERATE", "rec": "Expect LOW to MODERATE swings in this regime."},
            "high": {"color": "red", "desc": "HIGH to EXTREME", "rec": "Stay alert for HIGH to EXTREME swings in this regime."},
        }

        self.epoch = time.time()    #for output path


    @classmethod
    def instantiable(cls):
        return False

    def get_handlers(self, spec):
        """
        Define various callbacks that take a dataframe, spec
        and compute.
        """
        return {}

    ###########################################
    # Process the change points detection spec
    ###########################################
    def process_spec(self, spec, data):
        msg = ""

        if data is None:
            msg = "No dataset loaded" + "\n"
            logger.exception(
                f"Spec: {spec['name']} -- skipping",
                extra={"transform": self.name}
            )
            return None

        name = spec["name"]
        config = spec["config"]

        # use the specified strategy to detect change points
        detector = config['detector']
        method = detector.get('method', 'default')
        if method == "default":
            # we use the default strategy
            r = self.process_spec_default(spec, data)
        else:
            # Custom callback
            r = self.process_spec_custom(spec, data)

        # The spec processor can return multiple dataframes
        result = {name: r}

        logger.debug(
            f"Processed changepoint detection",
            extra={"transform": self.name, "data": json.dumps(result, indent=4, cls=SafeEncoder)}
        )

        return result

    def process_spec_custom(self, spec, data):
        msg = ""

        name = spec["name"]

        # Custom detector...
        detector = config['detector']
        method = detector['method']

        if hasattr(self, method):
            handler = getattr(self, method)
            df = handler(data)
        else:
            logger.exception(
                f"No method defined to detect changepoints",
                extra={"transform": self.name, "data": json.dumps(detector, indent=4)}
            )
            df = None

        return df

    def process_spec_default(self, spec, data):
        """
        Run the default change point detection
        """

        ## note: we need to save viz objects immediately after creation
        ## or they will be overwritten when the next viz object is created

        msg = ""
        name = spec['name']

        indicator_ts = data['indicator']
        cpd = self.run_changepoint_detector(indicator_ts, name)
        changepoints = cpd['changepoints']

        # save changepoint visualization
        viz = cpd['viz']
        filename = f"{name}-changepoints.png"
        l_msg = self.store_viz(spec, filename, viz)
        msg += l_msg
        cpd.pop('viz', None)

        observations = {}
        for observation, observation_ts in data['observations'].items():
            regimes = self.compute_regimes(indicator_ts, observation_ts, changepoints)
            viz = self.visualize_regimes(observation_ts, regimes)
            observations[observation] = {
                "regimes": regimes,
                "viz": viz
            }
            # save regimes visualization
            filename = f"{name}-regime-{observation}.png"
            l_msg = self.store_viz(spec, filename, viz)
            msg += l_msg
            observations[observation].pop('viz', None)

        logger.debug(
            f"Saved visualizations",
            extra={"transform": self.name, "data": msg}
        )

        result = {
            "changepoints": cpd,
            "observations": observations
        }

        return result


    def run_changepoint_detector(self, df, name):
        # init the change point detector object
        cpd_obj = BaseChangePointDetectorModel(df)

        # run the change point detection algorithm
        changepoints = cpd_obj.detect_changepoints(method="hybrid")

        # get the viz
        viz = cpd_obj.visualize_changepoints(df, changepoints, name)

        cpd = {
            "changepoints": changepoints,
            "viz": viz
        }

        return cpd

    def fit_project_linear_regression(self, y):
        x = np.array(range(0, len(y)))
        y_min = min(y)

        # shift to bring y_min to 0
        ynorm = (y - y_min)

        # fit the polynomial and compute coefficients
        z = np.polyfit(x, ynorm, 1)

        # get the projection as per the fit to data
        y_proj = np.polyval(z, len(y)-1) + y_min

        return y_proj

    def compute_impact_factor(self, data, start, end):
        y = data[start:end]
        y_mean = np.mean(y)

        y_proj = self.fit_project_linear_regression(y)

        # expected lift factor
        lfactor_EXP = y_proj/y_mean

        # observed lift factor
        lfactor_OBS = data[end]/y_mean

        # impact
        if lfactor_EXP < 1 and lfactor_OBS < 1:
            impact = (lfactor_EXP) / (lfactor_OBS)
        elif lfactor_EXP == 1:
            impact = 0   # If the expected lift factor is 1 i.e there is no change before and after cp, impact is 0
        else:
            impact = (lfactor_OBS-1) / (lfactor_EXP-1)

        direction = lfactor_OBS / lfactor_EXP
        if direction < 1 and impact != 0 :
            impact = 1 / impact

        return abs(impact)


    def compute_regimes(self, indicator_ts, observation_ts, changepoints):
        short_regime_range = 7
        damping = 0.1
        max_cap_factor = 1.5

        dates = observation_ts.index.values
        indicator_ts = indicator_ts.values
        observation_ts = observation_ts.values

        changepoint_indxs = [0] + changepoints

        regime_data = []
        for i in range(0, len(changepoint_indxs)-2):
            ## compute the expected impact to value in the next regime

            # get start and end indices for current regime
            start = changepoint_indxs[i]
            end = changepoint_indxs[i+1]

            # get prev regime start indices
            prev_start = changepoint_indxs[max(i-1,0)]

            # project what the ts value could have been at the end of current regime
            ts_values = observation_ts[start:end]
            ts_values_proj = self.fit_project_linear_regression(ts_values)

            # compute the impact factor at the change point for current regime
            impact_factor = self.compute_impact_factor(indicator_ts, start, end)

            # cmopute the caution indicator
            caution = self.compute_caution(impact_factor)

            # if very short current regime use prev two regimes
            if end-start <= short_regime_range:
                start = prev_start
                ts_values = observation_ts[start:end]

            # get the ts values over last 30 day time period to account for month-end highs
            ts_values_month = observation_ts[max(0,end-30):end]

            # get prev, next regime dates
            next_regime_dates = dates[changepoint_indxs[i+1]:changepoint_indxs[i+2]]
            prev_regime_dates = dates[start:end]
            prev_regime_len   = len(prev_regime_dates)-1


            # compute potential max ts value in the next regime
            if impact_factor >= 1:
                m1 = max(ts_values)
                m2 = min(ts_values_proj*math.sqrt(impact_factor), max(ts_values)*(impact_factor**damping))
                m3 = min(np.median(ts_values)*math.sqrt(impact_factor), max(ts_values)*(impact_factor**damping))
                max_ts_values = [round(x) for x in sorted([m1, m2, m3], reverse=True)]
                if caution == 'low':
                    max_cap = np.median(ts_values) * max_cap_factor
                    max_ts_values = [round(min(x, max_cap)) for x in max_ts_values]
                # we have to consider the last month high
                max_ts_values = sorted(max_ts_values + [round(max(ts_values_month))], reverse=True)[0:3]

                min_ts_values = [round(x) for x in sorted([min(ts_values), min(ts_values), min(ts_values)])]
            else:
                max_ts_values = [round(x) for x in sorted([max(ts_values), max(ts_values), max(ts_values)])]

                m1 = min(ts_values_month) #min(ts_values)
                m2 = ts_values_proj * math.sqrt(impact_factor)
                m3 = np.mean(ts_values) * math.sqrt(impact_factor)
                min_ts_values = [round(x) for x in sorted([m1, m2, m3])]

            # generate the context string
            rd = {
                "start_date": next_regime_dates[0],
                "end_date": next_regime_dates[-1],
                "prev_regime_start_date": prev_regime_dates[0],
                "prev_regime_end_date": prev_regime_dates[-1],
                "impact_factor": impact_factor,
                "min_value": min_ts_values,
                "max_value": max_ts_values,
                "prev_max_value": max(ts_values),
            }
            context = self.generate_regime_context(rd)
            rd['context'] = context

            regime_data.append(rd)

        # add the special case of final regime
        context = self.generate_regime_context(rd, final_rec=True)
        rd['context'] = context
        regime_data[-1] = rd

        return regime_data

    def compute_caution(self, impact_factor):
        caution = 'low'
        if impact_factor > 4 and impact_factor <= 6:
            caution = "medium"
        if impact_factor > 6:
            caution = "high"
        return caution


    def generate_regime_context(self, rd, final_rec=False):

        def millify(n):
            n = float(n)
            millnames = ['','K','M','B']
            millidx = max(0,min(len(millnames)-1,
                                int(math.floor(0 if n == 0 else math.log10(abs(n))/3))))
            return str(round(n / 10**(3 * millidx), 2)) + str(millnames[millidx])

        impact_factor = rd['impact_factor']
        caution = self.compute_caution(impact_factor)

        if caution == 'low':
            high_str = f"around ${millify(rd['max_value'][0])}"
        if caution == 'medium':
            high_str = f"between ${millify(rd['max_value'][1])} and ${millify(rd['max_value'][0])}"
        if caution == 'high':
            high_str = f"between ${millify(rd['max_value'][2])} and ${millify(rd['max_value'][1])} in the near future"

        multiple_str = round(rd['max_value'][0] / rd['prev_max_value'], 2)

        regime_start_date       = pd.to_datetime(str(rd['start_date'])).strftime("%-d %B, %Y")
        prev_regime_start_date  = pd.to_datetime(str(rd['prev_regime_start_date'])).strftime("%-d %B, %Y")
        prev_regime_end_date    = pd.to_datetime(str(rd['prev_regime_end_date'])).strftime("%-d %B, %Y")

        context = [f"{regime_start_date}: Regime change detected."] + \
                    [f"Expecting {self.cautions[caution]['desc']} change in spikiness from previous regime spanning {prev_regime_start_date} to {prev_regime_end_date}."] + \
                    [f"Recommendation 1: Expect highs around {multiple_str}x of previous regime high."] + \
                    [f"Recommendation 2: Expect highs {high_str}."]

        if final_rec:
            context[1] = f"{self.cautions[caution]['rec']}"

        return " ".join(context)


    def visualize_regimes(self, observation_ts, regime_data):
        fig, ax = plt.subplots(figsize=(15, 7))

        ax.plot_date(observation_ts.index, observation_ts.values,
                     linestyle='-', marker='')

        markers = {
            "^": {"color": "blue"},
            "v": {"color": "red"},
        }

        max_ts_value = 0
        sgn = 1

        # for each regime to visualize
        for rd in regime_data:

            impact_factor = rd['impact_factor']
            x = mdates.date2num(rd['start_date'])

            caution = self.compute_caution(impact_factor)

            # Create rectangle patches for the projection bands
            for i in range(0, len(rd['max_value'])):
                height = rd['max_value'][i]-rd['min_value'][i]
                rect = Rectangle((x, rd['min_value'][i]),
                                 width=mdates.date2num(rd['end_date'])-mdates.date2num(rd['start_date']),
                                 height=height,
                                 linewidth=1,
                                 edgecolor='r',
                                 facecolor=self.cautions[caution]['color'],
                                 alpha=0.1)

                # Add the patch to the Axes
                ax.add_patch(rect)

                # check for exit
                if caution == 'low' and i == 0:
                    break
                if caution == 'medium' and i == 1:
                    break

            # add a vertical line for the regime boundary
            ax.axvline(x,
                       linestyle='--',
                       color='black')

            dt = pd.to_datetime(str(rd['start_date'])).strftime("%Y-%m-%d")
            y = observation_ts[observation_ts.index==dt].values[0]
            regime_start_date = pd.to_datetime(str(rd['start_date'])).strftime("%-d %B, %Y")
            ax.annotate(regime_start_date,
                        (x, y),
                        xytext=(2, sgn*15),
                        xycoords='data',
                        textcoords='offset points',
                        bbox=dict(boxstyle="round", fc="white", ec="gray", alpha=0.75))

            marker = '^' if impact_factor >= 1 else 'v'
            ax.plot(x, y,
                    marker=marker,
                    color=markers[marker]['color'],
                    markersize=12)

            # tracking vars
            max_ts_value = max_ts_value if max(rd['max_value'])<max_ts_value else max(rd['max_value'])
            sgn *= -1

        ax.set_ylim(bottom=0, top=max_ts_value)
        ax.set_ylabel('Value')
        ax.set_title('Projected Regime Bands')

        fig.autofmt_xdate()
        return plt


    ###########################################
    # Helper Functions
    ###########################################

    def update_frame(self, source, description, df, lineage=None):
        if isinstance(source, str):
            name = source
        else:
            name = source["name"]

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    ###########################################
    # I/O Functions
    ###########################################

    def read_s3_data(self, filename, params={}):
        # assume we have a resolved s3fs object
        s3 = self.args['s3']
        if s3.exists(filename):
            try:
                df = pd.read_csv(s3.open(filename), **params)
                return df
            except:
                pass
        return None

    def get_dataset_s3(self, spec, source, paths, start_date, end_date):
        '''
        Gets all files from paths and puts them together
        into a single dataframe. If self.args['cache']==True,
        then this consolidated dataframe is cached / read from cache
        as applicable.
        '''
        msg = ""

        run_date    = self.args['run_date']
        config      = spec['config']
        dataset     = source['dataset']
        params      = source.get('params', {})

        cache = self.args.get("cache", False)
        cachename = f"{dataset}-{start_date}-to-{end_date}"
        cachefile = f"cache/{self.name}-rawdata-cache-" + cachename + ".csv"

        # read from cache if available
        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                msg = f"Location: {cachefile}" + "\n"
                df = pd.read_csv(cachefile, **params)
                logger.debug(f"Read cached {dataset}", extra={"transform": self.name, "data": msg})
                return df

        # read from S3
        dfs = []
        for path in paths:
            _df = self.read_s3_data(path, params)
            if _df is None:
                msg += f"Path error, skipping: {path}" + "\n"
                continue
            msg += f"Read from path: {path}" + "\n"
            dfs.append(_df)
        df = pd.concat(dfs)

        logger.debug(f"Read fresh {dataset}", extra={"transform": self.name})

        # Cache it for future use
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if (len(paths) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": paths,
                    },
                ],
            }

        if not self.state.has_frame(spec['name']):
            self.update_frame(spec, f"Dataset: {dataset}", df, lineage)

        return df

    def load_dataset(self, spec, name, source, datewindow):
        msg = ""

        source_id       = source['source_id']
        dataset         = source['dataset']
        source_version  = source.get('source_version', 'v1')

        start_date      = datewindow['start_date']
        end_date        = datewindow['end_date']

        if source_id == "custom":
            # we have a custom defined method in the derived class to generate the dataset
            if not hasattr(self, dataset):
                logger.exception(
                    f"No handler for: {dataset}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
                return None

            handler = getattr(self, dataset)
            params = source.get("params", {})
            df = handler(start_date, end_date, params)
            if df is None:
                logger.exception(
                    f"Couldn't load dataset for {spec['name']}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
                return None

            msg += f"Loaded using custom method: {dataset}" + "\n"
        else:
            # we are using the SDK to get the dataset
            datacred = self.args['datacred']
            doodle = TransformDoodle(self, self.state, datacred)

            d_source, paths = doodle.get_source_paths(start=start_date,
                                                    end=end_date,
                                                    name=dataset,
                                                    version=source_version,
                                                    source_id=source_id)

            msg += f"Paths: {paths}" + "\n"

            # Insert a read action..
            try:
                result = doodle.update_source(d_source['id'], {})
                logger.debug(f"Updated Doodle: {dataset}",
                             extra={
                                 'transform': self.name,
                                 'data': json.dumps(result, indent=4, cls=SafeEncoder)
                             })
            except:
                logger.exception(f"Unable to update Doodle {dataset}",
                                 extra={
                                     'transform': self.name
                                 })

            # read the data from S3
            df = self.get_dataset_s3(spec, source, paths, start_date, end_date)


        # check if we need to perform a post-processing step on the input data
        postprocess = source.get("postprocess")
        if postprocess != None:
            if not hasattr(self, postprocess['method']):
                logger.exception(
                    f"No post-process handler for: {dataset}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
            else:
                handler = getattr(self, postprocess['method'])
                params = postprocess.get("params", {})
                df = handler(df, params)
                msg += f"Post-processed dataset={dataset} using handler: {postprocess}" + "\n"

        msg += note(df, f"Input Dataset: {dataset}")
        logger.debug(
            f"Loaded dataset={dataset} for source={name}",
            extra={"transform": self.name, "data": msg}
        )

        return df

    def store_viz(self, spec, filename, viz):
        appname     = spec.get('app', self.name)
        name        = spec['name']
        namespace   = spec.get('namespace', 'default')
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = self.epoch

        msg = ""

        # first, check if we need to store locally
        if self.args.get('testmode', False):

            outputpath = self.args['test_outpath']
            try:
                os.makedirs(outputpath)
            except:
                pass

            testplotfile = os.path.join(outputpath, filename)
            viz.savefig(testplotfile)

            msg += f"Stored (local): {testplotfile}" + "\n"

        # write to s3
        # where are we storing it?
        targetdir = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")
        vizfile = os.path.join(targetdir, f"{filename}")

        img_data = io.BytesIO()
        viz.savefig(img_data, format='png')
        img_data.seek(0)
        with s3.open(vizfile, 'wb') as fd:
            fd.write(img_data.getbuffer())

        msg += f"Stored (remote): {vizfile}" + "\n"

        return msg


    def s3_store_result(self, spec, result):
        appname     = spec.get('app',self.name)
        name        = spec['name']
        namespace   = spec.get('namespace', 'default')
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = self.epoch


        # where are we storing it?
        targetdir = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")

        resultfile = os.path.join(targetdir, f"data.json")
        metadatafile = os.path.join(targetdir, f"metadata.json")

        metadata = self.get_default_metadata(self.state)
        metadata['spec'] = spec

        # write to s3
        with s3.open(resultfile, 'w') as fd:
            json.dump(result, fd, indent=4, cls=SafeEncoder)
        with s3.open(metadatafile, 'w') as fd:
            json.dump(metadata, fd, indent=4, cls=SafeEncoder)

        msg = f"s3 location: {resultfile}" + "\n"
        msg += f"metadata location: {metadatafile}" + "\n"

        logger.debug(f"Wrote change point result to S3",
                        extra={"transform": self.name,
                                "data": msg})



    def db_store_result(self, spec, result):
        name    = spec['name']
        config  = spec['config']
        store   = config['store']

        testmode = self.args.get('testmode', False)
        env      = 'test' if testmode else 'prod'

        # Get the db engine
        engine_name = store['params'][env]['engine']
        engine = self.engines[engine_name]

        # write results to db
        table_name = store['params'][env]['table']
        ret = results.to_sql(table_name,
                            engine,
                            if_exists='append',
                            index=False)
        # write data to db
        data_table_name = name.replace('-','_')
        ret = data.to_sql(data_table_name,
                            engine,
                            if_exists='append',
                            index=False)

        msg = f"DB engine: {engine}" + "\n"

        logger.debug(f"Wrote monitor results to DB",
                        extra={"transform": self.name,
                                "data": msg})

    def store_result(self, spec, result):
        name    = spec['name']
        config  = spec['config']
        store   = config.get('store', {"sink": "s3"})

        for f in ["sink"]:
            if f not in store:
                logger.exception(
                    f"Store has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                return

        sink = store['sink']
        if sink == "s3":
            # store in s3
            self.s3_store_result(spec, result)
        elif sink == "db":
            # store in db
            self.db_store_result(spec, result)
        else:
            logger.exception(f"Unknown store for dataset: {name}",
                         extra={
                             'transform': self.name
                         })


    def get_datewindow(self, source, spec):
        datewindow = {}
        default_delta = 180

        start_date_str = source.get('start_date')
        end_date_str = source.get('end_date')

        if not end_date_str:
            logger.debug(f"end date not specified in {spec['name']}. Using default as yesterday's date.")
            end  = self.args['run_date'] # yesterday
            end_date_str = end.isoformat()

        if not start_date_str:
            logger.debug(f"start date not specified in {spec['name']}. Using default as 180 days prior to end date. ")
            start = dateparser.parse(end_date_str).date() + timedelta(days=-default_delta)
            start_date_str = start.isoformat()

       # If start and end dates are specified, check if date is parseable
        try:
            start_date = dateparser.parse(start_date_str).date()
            end_date = dateparser.parse(end_date_str).date()
        except:
            logger.exception(
                        f"Invalid start date or end date. Skipping the spec {spec['name']}.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
            datewindow = None # be explicit
            return datewindow

        if start_date > end_date:
            logger.exception(
                        f"Start date greater than end date. Skipping the spec {spec['name']}.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
            datewindow = None # be explicit
            return datewindow

        datewindow['start_date'] = start_date
        datewindow['end_date'] = end_date

        return datewindow

    ###########################################
    # Expectation Testing Functions
    ###########################################

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "Start execution", extra=self.config.get_extra({"transform": self.name})
        )

        # Will be used in other places..
        self.state = state

        # Get the profile spec
        is_valid, profile, msg = profilespec.get_profile(self, "policyapp.changepoint")
        if is_valid:
            logger.debug(
                f"Loaded profilespec",
                extra={"transform": self.name, "data": msg}
            )
        else:
            logger.error(
                f"Could not load profilespec",
                extra={"transform": self.name, "data": msg}
            )
            raise Exception("could not load profilespec")

        specs = profile.get("specs", None)
        if specs is None:
            raise Exception("Could not find 'specs' in profile")

        # Now go through each spec and process it
        for spec in specs:

            ## first, some checks on the spec
            do_process_spec = True
            name = spec.get('name', 'NO_SPEC_NAME')

            enabled = spec.get("active", True)
            if not enabled:
                logger.debug(
                    f"Spec not enabled, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                do_process_spec = False
                continue

            for f in ["name", "config"]:
                if f not in spec:
                    logger.exception(
                        f"Spec has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue

            config = spec['config']

            for f in ["source", "detector"]:
                if f not in config:
                    logger.exception(
                        f"Spec config has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue

            ## we can now proceed with processing the spec
            source = config['source']
            for f in ["indicator", "observations"]:
                if f not in source:
                    logger.exception(
                        f"Spec config has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            # expections have occured. continue to next spec
            if do_process_spec == False:
                continue

            # get time_window for indicator and observation data set
            datewindow = self.get_datewindow(source, spec)
            if datewindow is None :
                do_process_spec = False
                continue

            # first, load the indicator dataset
            data = {}
            data['indicator'] = self.load_dataset(spec, 'indicator', source['indicator'], datewindow)
            if data['indicator'] is None:
                do_process_spec = False
                continue

            # then, load all the observations datasets
            data['observations'] = {}
            for ob_dataset, dataset in source['observations'].items():
                data['observations'][ob_dataset] = self.load_dataset(spec, ob_dataset, dataset, datewindow)

            # then, process it
            result = self.process_spec(spec, data)
            if result is None:
                continue

            ## store the expectation validation result
            self.store_result(spec, result)


        # Done
        logger.debug(
            "Complete execution", extra=self.config.get_extra({"transform": self.name})
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass
